#!/usr/bin/env python

# -*- coding: utf-8 -*-

from __future__ import absolute_import

from __future__ import print_function

from __future__ import division

# -----------------------------------------------------------------------------

from wsurf.geolab.plotter.plotmanager import PlotManager

from wsurf.geolab.gui.geolabbsplinesurface import GeolabBSplineSurface

# -----------------------------------------------------------------------------

'''plotmanager.py: The scene manager class'''

__author__ = 'Davide Pellis'

# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
#                                Scene Manager
# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------


class SceneManager(PlotManager):

    def __init__(self, scene_model=None):
        PlotManager.__init__(self, scene_model)

        self._objects = {}

        self._selected = None

        self._last_object = None

        self._counter = 0

        self._handler = None

        self._geometry_change_callbacks = []

        self._object_changed_callbacks = []

        self._object_change_callbacks = []

        self._object_save_callbacks = []

        self._object_saved_callbacks = []

        self._object_added_callbacks = []

        self._object_add_callbacks = []

    # -------------------------------------------------------------------------
    #                                Properties
    # -------------------------------------------------------------------------

    @property
    def scene_model(self):
        return self._scene_model

    @scene_model.setter
    def scene_model(self, scene_model):
        self._scene_model = scene_model
        self._scene = scene_model.mayavi_scene
        for obj in self._objects:
            obj.scene_model = scene_model

    @property
    def current_object(self):
        if self._selected is None:
            try:
                return self._objects[self._objects.keys()[0]]
            except:
                return None
        try:
            return self._objects[self._selected]
        except:
            self._selected = None
            return self.current_object

    @property
    def objects(self):
        return self._objects[self._selected]

    @property
    def last_object(self):
        if self._last_object is None:
            try:
                return self._objects[self._objects.keys()[0]]
            except:
                return None
        try:
            return self._objects[self._last_object]
        except:
            self._last_object = None
            return self.last_object

    @property
    def engine(self):
        return self._scene_model.engine

    #--------------------------------------------------------------------------

    def add_callback(self, key, callback):
        if key == 'object_change':
            self._object_change_callbacks.append(callback)
        elif key == 'object_changed':
            self._object_changed_callbacks.append(callback)
        elif key == 'object_save':
            self._object_save_callbacks.append(callback)
        elif key == 'object_saved':
            self._object_saved_callbacks.append(callback)
        elif key == 'object_added':
            self._object_added_callbacks.append(callback)
        elif key == 'object_add':
            self._object_add_callbacks.append(callback)

    def object_changed(self):
        for callback in self._object_changed_callbacks:
            try:
                callback()
            except:
                pass

    def object_change(self):
        for callback in self._object_change_callbacks:
            try:
                callback()
            except:
                pass

    def object_save(self, file_name):
        for callback in self._object_save_callbacks:
            try:
                callback(file_name)
            except:
                pass

    def object_saved(self, file_name):
        for callback in self._object_saved_callbacks:
            try:
                callback(file_name)
            except:
                pass

    def object_added(self):
        for callback in self._object_added_callbacks:
            try:
                callback()
            except:
                pass

    def object_add(self):
        for callback in self._object_add_callbacks:
            try:
                callback()
            except:
                pass

    # -------------------------------------------------------------------------
    #                            Get from pipeline
    # -------------------------------------------------------------------------

    def update(self, **kwargs):
        obj = kwargs.get('object', None)
        if obj is not None:
            self._objects[obj].update(**kwargs)
        else:
            PlotManager.update(self, **kwargs)

    def clear(self):
        for key in self._objects:
            self._objects[key].clear()
        PlotManager.clear(self)

    def clear_scene(self):
        for scene in self.engine.scenes:
           if scene != self.scene_model:
               scene.remove()

    def add_object(self, geometry, name=None):
        self.object_add()
        if geometry.type == 'Bspline_Surface':
            GP = GeolabBSplineSurface(self.scene_model)
            GP.geometry = geometry
            obj = GP
        if name is not None:
            obj.name = name
        else:
            obj.name = 'obj_{}'.format(self._counter)
            self._counter += 1
        if len(self._objects) == 0:
            self._selected = obj.name
        self._last_object = obj.name
        self.remove(obj.name)
        self._objects[obj.name] = obj
        obj.scene_model = self.scene_model
        self.object_added()

    def remove(self, names=None):
        if names is None:
            if self._selected is None:
                names = [self._objects.keys()[0]]
            else:
                names = [str(self._selected)]
        if type(names) is str:
            names = [names]
        for key in names:
            try:
                self._objects[key].clear()
                del self._objects[key]
                self._selected = None
            except:
                pass
        PlotManager.remove(self, names)

    def core_plot(self):
        PlotManager.core_plot(self)
        for key in self._objects:
            obj = self._objects[key]
            obj.update_plot()

    def hide(self, **kwargs):
        name = kwargs.get('name', None)
        if name is None:
            super(SceneManager, self).hide()
            for key in self._objects:
                try:
                    self._objects[key].hide()
                except:
                    pass
        else:
            obj = kwargs.get('object', None)
            if obj is not None:
                self._objects[obj].hide(**kwargs)
            else:
                PlotManager.hide(self, **kwargs)

    def get_object(self, name):
        try:
            obj = self._objects[name]
        except:
            obj = None
        return obj

    # -------------------------------------------------------------------------
    #                                 Record
    # -------------------------------------------------------------------------

    def shoot(self, kink=False):
        for key in self._objects:
            obj = self._objects[key]
            obj.shoot(kink)

    def interpolate_frames(self, N, key_frames_number):
        for key in self._objects:
            obj = self._objects[key]
            obj.interpolate_frames(N, key_frames_number)

    def clear_records(self):
        for key in self._objects:
            obj = self._objects[key]
            obj.clear_records()

    def plot_frame(self, frame_number):
        for key in self._objects:
            obj = self._objects[key]
            obj.plot_frame(frame_number)

    def plot_key_frame(self, key_number):
        for key in self._objects:
            obj = self._objects[key]
            obj.plot_key_frame(key_number)

    # -------------------------------------------------------------------------
    #                                  Selection
    # -------------------------------------------------------------------------

    def select_object(self):
        if self.current_object == None:
            return
        self.current_object.highlight()
        ranges = {}
        count = 0
        for key in self._objects:
            obj = self._objects[key]
            N = obj.selection_on()
            ranges[key] = [count, count+N]
            count += N
        def picker_callback(picker):
            pick_id = picker.cell_id
            if pick_id >= 0:
                for key in self._objects:
                    try:
                        obj = self._objects[key]
                        if picker.actor in obj.object_selection_actors:
                            if key != self._selected:
                                self.current_object.highlight_off()
                                self.object_change()
                                self._selected = key
                                self.object_changed()
                                self.current_object.highlight()
                    except IndexError:
                        pass
        s = self.engine.current_scene
        p = s._mouse_pick_dispatcher.callbacks
        if len(p) == 0:
            s.on_mouse_pick(picker_callback,'cell')
        else:
            p[0] = (picker_callback,'cell','Left')
        a = s._mouse_pick_dispatcher._active_pickers['cell']
        a.tolerance = self.picker_tolerance


    def select_off(self):
        for key in self._objects:
            obj = self._objects[key]
            obj.select_off()
            try:
                obj.hilight_off()
            except:
                pass
        scene = self.engine.current_scene
        p = scene._mouse_pick_dispatcher.callbacks
        def picker_callback(picker):
            return
        for i in range(len(p)):
            p[i] = (picker_callback,'cell','Left')


#!/usr/bin/env python

# -*- coding: utf-8 -*-

from __future__ import absolute_import

from __future__ import print_function

from __future__ import division

from tvtk.api import tvtk

import numpy as np

# ------------------------------------------------------------------------------

from wsurf.geolab.plotter.plotmanager import PlotManager

from wsurf.geolab.plotter.edgesource import Edges

from wsurf.geolab.plotter.pointsource import Points

from wsurf.geolab.plotter.facesource import Faces

# ------------------------------------------------------------------------------

'''-'''

__author__ = 'Davide Pellis'


# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
#                               Mesh PlotManager
# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------


class BsplineSurfacePlotManager(PlotManager):

    def __init__(self, scene_model=None):
        PlotManager.__init__(self, scene_model)

        self.scale = 1

        self.glyph_scale = 1.5

        self.vector_scale = 1

        self.selected_vertices = []

        self.vertex_color = 'cornflower'

        self.control_color = (10, 10, 10)

        self.surface_color = 'cornflower'

        self.selection_color = 'yellow'

        self.view_mode = 'solid'

        self._show_virtual = False

        self._r = None

        self._g = None

        self._bspline = None

        self._face_sources = []

        self._edge_sources = []

        self._updating = True

        self.__dimensions = (0, 0, 0, 0)

        self.__clear = False

    # -------------------------------------------------------------------------
    #                              Properties
    # -------------------------------------------------------------------------

    @property
    def type(self):
        return 'Bspline_surface_plot_manager'

    @property
    def bspline(self):
        return self._bspline

    @bspline.setter
    def bspline(self, bspline):
        if bspline.type != 'Bspline_Surface':
            raise ValueError('*bspline* attribute must be a B-Spline_Surface!')
        self._bspline = bspline
        self._set_r()
        self._set_g()
        self.vertex_data = np.zeros(bspline.S)

    @property
    def geometry(self):
        return self._bspline

    @geometry.setter
    def geometry(self, bspline):
        self.bspline = bspline

    @property
    def r(self):
        return self._r * self.scale

    @property
    def g(self):
        return self._g * self.glyph_scale

    @property
    def v(self):
        return 12 * self._r * self.vector_scale

    @property
    def updating(self):
        return self._updating

    @updating.setter
    def updating(self, updating):
        if type(updating) == bool:
            self._updating = updating
        else:
            raise ValueError('the updating attribute must be a boolean!')

    @property
    def vertex_data(self):
        if self._vertex_data is None:
            self._vertex_data = np.zeros(self.bspline.S)
        return self._vertex_data

    @vertex_data.setter
    def vertex_data(self, vertex_data):
        self._vertex_data = vertex_data

    @property
    def object_selection_actors(self):
        return self.get_actor('virtual-bspline').actors

    # -------------------------------------------------------------------------
    #                              Size set
    # -------------------------------------------------------------------------

    def update_size(self):
        self._set_r()
        self._set_g()

    def check_updates(self):
        updated = False
        if self.__dimensions[0] != self.bspline.Cu:
            updated = True
        if self.__dimensions[1] != self.bspline.Cv:
            updated = True
        if self.__dimensions[2] != self.bspline.u_sampling:
            updated = True
        if self.__dimensions[3] != self.bspline.v_sampling:
            updated = True
        if updated:
            # self._set_r()
            # self._set_g()
            self._vertex_data = None
            self.__clear = True
        self.__dimensions = (self.bspline.Cu, self.bspline.Cv,
                             self.bspline.u_sampling, self.bspline.v_sampling)
        return updated

    def _set_r(self):
        if self.bspline is None or self.bspline.C == 0:
            self._r = None
        else:
            box = (np.max(self.bspline.control_points, axis=0) -
                   np.min(self.bspline.control_points, axis=0))
            r = np.linalg.norm(box) / 800
            self._r = r

    def _set_g(self):
        if self.bspline is None or self.bspline.C == 0:
            self._g = None
        else:
            g = self.r
            self._g = g

    # -------------------------------------------------------------------------
    #                                Clear
    # -------------------------------------------------------------------------

    def clear(self, delete=True):
        PlotManager.clear(self)
        # self.picker_off()
        # self.remove_widgets()
        # self.selected_vertices = []

    # -------------------------------------------------------------------------
    #                            Plot functions
    # -------------------------------------------------------------------------

    def remove(self, names):
        PlotManager.remove(self, names)
        if type(names) is str:
            names = [names]
        for name in names:
            if name in self._face_sources:
                self._face_sources.remove(name)
            elif name in self._edge_sources:
                self._edge_sources.remove(name)

    def remove_face_sources(self):
        for name in self._face_sources:
            PlotManager.remove(self, name)
        self._face_sources = []

    def remove_edge_sources(self):
        for name in self._edge_sources:
            PlotManager.remove(self, name)
        self._edge_sources = []

    def clear_check(self, name):
        clear = False
        if name not in self.sources:
            clear = True
        if not self.updating:
            clear = True
        if self.__clear:
            clear = True
        return clear

    def start_plot(self):
        PlotManager.start_plot(self)
        self.check_updates()

    def core_plot(self, **kwargs):
        PlotManager.core_plot(self)

    def end_plot(self):
        self.plot_selected_vertices()
        self.__clear = False
        PlotManager.end_plot(self)

    def initialize_plot(self):
        pass
        # self.plot_edges()

    def plot_surface(self, **kwargs):
        if 'name' not in kwargs:
            kwargs['name'] = 'surface'
        else:
            self._face_sources.append(kwargs['name'])
        if 'color' not in kwargs:
            kwargs['color'] = self.surface_color
        if self.clear_check(kwargs['name']):
            self.bspline.plot_mode = 'surface'
            F = Faces(self.bspline, **kwargs)
            self.add(F)
            self._face_sources.append(kwargs['name'])
        else:
            self.bspline.plot_mode = 'surface'
            self.update(**kwargs)

    def plot_control_polygon(self, **kwargs):
        if 'update' not in kwargs:
            kwargs['update'] = True
        if 'name' not in kwargs:
            kwargs['name'] = 'control_polygon'
        else:
            self._edge_sources.append(kwargs['name'])
        if 'tube_radius' not in kwargs:
            if self.view_mode == 'wireframe':
                tube_radius = None
            elif self.view_mode == '3d':
                tube_radius = None
            else:
                tube_radius = self.r
            kwargs['tube_radius'] = tube_radius
        if 'color' not in kwargs:
            kwargs['color'] = self.control_color
        if 'glossy' not in kwargs:
            kwargs['glossy'] = 0.1
        if self.clear_check(kwargs['name']):
            self.bspline.plot_mode = 'control'
            M = Edges(self.bspline, **kwargs)
            self.add(M)
            self._edge_sources.append(kwargs['name'])
        else:
            self.bspline.plot_mode = 'control'
            self.update(**kwargs)

    def plot_vertices(self, **kwargs):
        if 'name' not in kwargs:
            kwargs['name'] = 'vertices'
        if 'radius' not in kwargs:
            kwargs['radius'] = self.r * 2.5
        if 'color' not in kwargs:
            kwargs['color'] = self.vertex_color
        if self.clear_check(kwargs['name']):
            P = Points(self.spline.control_points, **kwargs)
            self.add(P)
        else:
            self.update(**kwargs)

    def plot_glyph(self, **kwargs):
        if 'name' not in kwargs:
            kwargs['name'] = 'glyph'
        if 'points' not in kwargs:
            kwargs['points'] = self.bspline.control_points
        if 'radius' not in kwargs:
            kwargs['radius'] = self.r * 2.5
        if 'color' not in kwargs:
            kwargs['color'] = self.vertex_color
        if self.clear_check(kwargs['name']):
            P = Points(**kwargs)
            self.add(P)
        else:
            self.update(**kwargs)

    def hide_control_polygon(self):
        self.hide('control_polygon')

    def remove_control_polygon(self):
        self.remove('control_polygon')

    def hide_surface(self):
        self.hide('surface')

    def remove_surface(self):
        self.remove('surface')
        for name in self._face_sources:
            self.remove(name)
        self._face_sources = []

    def hide_vertices(self):
        self.hide('vertices')

    def remove_vertices(self):
        self.remove('vertices')

    # -------------------------------------------------------------------------
    #                             Virtual plot
    # -------------------------------------------------------------------------

    def selection_on(self):
        self.bspline.plot_mode = 'control'
        M = Edges(self.bspline,
                  name='virtual-bspline',
                  color=(0.5, 0.5, 0.5),
                  opacity=0.001, )
        self.add(M, pickable=True)
        return self.bspline.E

    def selection_off(self):
        self.remove('virtual-bspline')

    def virtual_vertices_on(self):
        if True:  # self.clear_check('virtual-vertices'):
            opacity = 0.001
            if self._show_virtual:
                opacity = 1
            M = Points(self.bspline.control_points,
                       name='virtual-vertices',
                       glyph_type='cube',
                       radius=self.r * 3,
                       opacity=opacity)
            self.add(M, pickable=True)
        else:
            self.update(name='virtual-vertices', radius=self.r * 1.1)

    def virtual_vertices_off(self):
        self.hide('virtual-vertices')

    # --------------------------------------------------------------------------
    #                            Selection plot
    # --------------------------------------------------------------------------

    def highlight(self):
        self.bspline.plot_mode = 'control'
        tube_radius = self.g * 1.03
        M = Edges(self.bspline,
                  tube_radius=None,
                  line_width=5 * self.scale,
                  color=self.selection_color,
                  shading=False,
                  name='highlight')
        self.hide('control_polygon')
        self.add(M)

    def highlight_off(self):
        self.show('control_polygon')
        self.remove('highlight')

    def plot_selected_vertices(self):
        if len(self.selected_vertices) > 0:
            if self.clear_check('selected-vertices'):
                # if self.view_mode == 'wireframe' or self.view_mode == '3d':
                # glyph = 'wirefrane'
                # radius = None
                # elif self.view_mode == 'solid':
                radius = self.g * 2.52
                glyph = 'sphere'
                P = Points(self.bspline.control_points,
                           vertex_indices=self.selected_vertices,
                           color=self.selection_color,
                           radius=radius,
                           shading=False,
                           glyph_type=glyph,
                           name='selected-vertices',
                           resolution=18)
                self.add(P)
            else:
                self.update(name='selected-vertices',
                            vertex_indices=self.selected_vertices,
                            radius=self.g * 2.52)
        else:
            self.hide('selected-vertices')

    def hide_selected_vertices(self):
        self.hide('selected-vertices')

    # -------------------------------------------------------------------------
    #                               Selection
    # -------------------------------------------------------------------------

    def select_all_vertices(self):
        self.select_off()
        self.virtual_vertices_on()
        self.selected_vertices = list(range(self.bspline.C))
        self.plot_selected_vertices()

        def callback(p_id):
            if p_id == -1:
                return
            v = p_id // 6
            if v not in self.selected_vertices:
                self.selected_vertices.append(v)
            else:
                self.selected_vertices.remove(v)
            self.plot_selected_vertices()

        self.picker_callback(callback, mode='cell', name='virtual-vertices')

    def select_vertices(self):
        self.select_off()
        self.virtual_vertices_on()
        self.selected_vertices = []

        def callback(p_id):
            if p_id == -1:
                return
            v = p_id // 6
            if v not in self.selected_vertices:
                self.selected_vertices.append(v)
            else:
                self.selected_vertices.remove(v)
            self.plot_selected_vertices()

        self.picker_callback(callback, mode='cell', name='virtual-vertices')

    def select_boundary_vertices(self):
        self.select_off()
        self.virtual_vertices_on()
        self.selected_vertices = []

        def v_callback(p_id):
            if p_id == -1:
                return
            v = p_id // 6
            boundaries = self.bspline.boundary_control_curves()
            if len(boundaries) == 4:
                corners = [boundaries[0][0], boundaries[0][-1],
                           boundaries[1][0], boundaries[1][-1]]
            else:
                corners = []
            if v not in corners:
                if v not in self.selected_vertices:
                    for boundary in boundaries:
                        if int(v) in boundary:
                            self.selected_vertices.extend(boundary)
                    self.plot_selected_vertices()
                elif v in self.selected_vertices:
                    for boundary in boundaries:
                        if int(v) in boundary:
                            for w in boundary:
                                self.selected_vertices.remove(w)
                    self.plot_selected_vertices()

        self.picker_callback(v_callback, mode='cell', name='virtual-vertices')

    def select_boundary_vertices_off(self):
        self.select_vertices_off()

    def on_vertex_selection(self, vertex_callback):
        self.select_off()
        self.virtual_vertices_on()
        self.selected_vertices = []

        def callback(p_id):
            if p_id == -1:
                return
            v = p_id // 6
            self.selected_vertices = [v]
            if vertex_callback is not None:
                vertex_callback(v)
            self.virtual_vertices_on()

        self.picker_callback(callback, mode='cell', name='virtual-vertices')

    def select_vertices_off(self):
        self.virtual_vertices_off()
        self.hide_selected_vertices()
        self.picker_off()
        self.selected_vertices = []

    def clear_selection(self):
        self.hide_selected_vertices()
        self.selected_vertices = []
        self.virtual_vertices_off()
        self.remove_widgets()

    def select_off(self):
        self.select_vertices_off()
        self.move_vertex_off()
        self.move_vertices_off()

    # -------------------------------------------------------------------------
    #                               Moving
    # -------------------------------------------------------------------------

    def move_vertex(self, vertex_index, interaction_callback=None,
                    end_callback=None):
        self.remove_widgets()
        S = tvtk.SphereWidget()
        point = self.bspline.control_points[vertex_index]
        S.center = point
        S.radius = self.g * 2
        S.scale = False
        S.representation = 'surface'
        # S.handle_property.color = (0, 0, 0)
        # S.handle_property.opacity = 0.3
        # S.modified()
        P = Points(points=point,
                   radius=self.g * 2.52,
                   color=self.selection_color,
                   shading=False,
                   name='widget-point')
        self.add(P)
        self.virtual_vertices_on()

        def i_callback(obj, event):
            self.virtual_vertices_off()
            c = obj.GetCenter()
            center = np.array([c[0], c[1], c[2]])
            self.bspline.control_points[vertex_index, :] = center
            self.update(name='widget-point', points=center)
            if interaction_callback is not None:
                interaction_callback()

        S.add_observer("InteractionEvent", i_callback)

        def e_callback(obj, event):
            if end_callback is not None:
                end_callback()
            self.virtual_vertices_on()
            point = self.bspline.control_points[vertex_index]
            S.center = point
            self.update(name='widget-point', points=point)

        S.add_observer("EndInteractionEvent", e_callback)

        self.add_widget(S, name='vertex_handler')

    def move_vertex_off(self):
        self.remove_widgets()
        self.hide('widget-point')
        self.virtual_vertices_off()

    def move_vertices(self, interaction_callback=None, start_callback=None,
                      end_callback=None):
        self.select_off()
        self.virtual_vertices_on()

        def v_callback(p_id):
            self.hide_selected_vertices()
            if p_id == -1:
                return
            v = p_id // 6
            self.selected_vertices = [v]
            if start_callback is not None:
                try:
                    start_callback()
                except:
                    pass
            self.move_vertex(v, interaction_callback, end_callback)

        self.picker_callback(v_callback, mode='cell', name='virtual-vertices')

    def move_vertices_off(self):
        self.selected_vertices = []
        self.move_vertex_off()
        self.select_vertices_off()
        self.picker_off()



# -*- coding: utf-8 -*-


# !/usr/bin/env python

# -*- coding: utf-8 -*-

from __future__ import absolute_import

from __future__ import print_function

from __future__ import division

import numpy as np

from scipy import sparse

# -----------------------------------------------------------------------------

from .geolab.geometry.polyline import Polyline

from .geolab import optimization

from .geolab import plotter

from .geolab import fitting

from .geolab import utilities

# -----------------------------------------------------------------------------

__author__ = 'Davide Pellis'

'''
-------------------------------------------------------------------------------
-------------------------------------------------------------------------------
                        WEINGARTEN B-SPLINE OPTIMIZATION
-------------------------------------------------------------------------------
                                     2020
-------------------------------------------------------------------------------
-------------------------------------------------------------------------------
                                    Notes
-------------------------------------------------------------------------------
 indexing:
 C = control points
 S = sample points
-------------------------------------------------------------------------------
 unknows vector X:
 3C | 3S | 4S | 3S | 2S + 2S

                | N1              | N2                  | N3
 [x_C, y_C, z_C, Nx_S, Ny_S, Nz_S, E_S, F_S, G_S, det_S, L_S, M_S, N_S,

 | N4      | N5                                                   | N6
  K_S, H_S, Ku^2_S, KuKv_S, Kv^2_S, Hu^2_S, HuHv_S, Hv^2_S, DK, DH, a, b, c]
-------------------------------------------------------------------------------
 K = (LN - M^2) / (EG - F^2)
 H = (LG - 2MF + NE) / 2(EG - F^2)
-------------------------------------------------------------------------------
-------------------------------------------------------------------------------
                                  Constraints
-------------------------------------------------------------------------------
                      fi = Hi_jk X_j X_k + bi_j X_j - ci
-------------------------------------------------------------------------------
            H_ij = 2 Hi_jk X_k + bi_j;    r_i = Hi_jk X_j X_k - ci
-------------------------------------------------------------------------------
-------------------------------------------------------------------------------
'''

# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
#
# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------

Base = optimization.GuidedProjectionBase


class WsurfOptimizer(Base):
    _N1 = 0

    _N2 = 0

    _N3 = 0

    _N4 = 0

    _N5 = 0

    _N6 = 0

    _bspline = None

    _curves = []

    _reference = None

    _boundary_curves = None

    _fixed_points = np.array([], 'i')

    _handle = None

    _scale = 1

    __scale = 1

    clusters = None

    include_boundary = False

    adaptive_curve = True

    normalize_geometry = True

    linear_coefficients = (0, 0, 0)

    n_curve_control_points = 4

    straightening = 0.1

    n_clusters = 1

    cluster_coefficients = [1, 1, 1, 1, 1]

    def __init__(self, bspline=None):
        Base.__init__(self)

        weights = {

            'normals': 1,

            'first': 1,

            'determinant': 1,

            'second': 1,

            'gaussian': 1,

            'mean': 1,

            'delta': 1,

            'squared_gradients': 1,

            'gradients_norms': 1,

            # -----------------------------------------------------------------
            #                           User weights
            # -----------------------------------------------------------------

            'hk': 0,

            'reference': 0,

            'control_lengths': 0,

            'fixed': 100,

            'gliding': 0,

            'control_fairness': 0,

            'curvature_fairness': 0,

            'curvature_minimization': 0,

            'isolines_parallel': 0,

            'linear_fitting': 0,

            'curve_fitting': 0,

        }

        self.add_weights(weights)

        self.bspline = bspline

    # --------------------------------------------------------------------------
    #
    # --------------------------------------------------------------------------

    @property
    def curves(self):
        return self._curves

    @property
    def bspline(self):
        return self._bspline

    @bspline.setter
    def bspline(self, bspline):
        self._bspline = bspline
        if bspline is not None:
            self.normalize_bspline()
            self._fixed_points = np.array([], 'i')
            self._handle = None
            self.initialize = True
            self.add_value('curve_closest', None)

    @property
    def sampling(self):
        return self.bspline.sampling

    @sampling.setter
    def sampling(self, sampling):
        self.bspline.sampling = sampling
        self.reinitialize()

    @property
    def handle(self):
        if self._handle is None:
            return np.array([], 'i')
        else:
            return self._handle

    @handle.setter
    def handle(self, handle):
        if handle is None:
            handle = np.array([], 'i')
        self._handle = np.array(handle, 'i')

    @property
    def fixed_points(self):
        return self._fixed_points

    @property
    def scale(self):
        return self._scale

    @scale.setter
    def scale(self, scale):
        if scale > 0:
            self._scale = scale
            self.normalize_bspline()
            self.reinitialize()

    # -------------------------------------------------------------------------
    #                               Initialization
    # -------------------------------------------------------------------------

    def fix_points(self, points):
        points = np.array(points, 'i')
        self._fixed_points = np.unique(np.hstack((self.fixed_points, points)))

    def fix_boundary_points(self):
        self._fixed_points = self.bspline.boundary_control_points()

    def unfix_points(self, points):
        mask = np.invert(np.in1d(self._fixed_points, points))
        self._fixed_points = self._fixed_points[mask]

    def set_reference(self):
        self.add_value('reference', np.copy(self._bspline.control_points))
        self.add_value('control_lengths', self.control_lengths())
        self.add_value('control_diagonals', self.control_diagonals())
        self.make_boundary_curves()
        self.initialize = True

    def make_boundary_curves(self):
        bdr = self.bspline.boundary_control_curves()
        self._boundary_curves = []
        for b in bdr:
            P = self.bspline.control_points[b]
            crv = Polyline(P, closed=True)
            crv.refine(5)
            self._boundary_curves.append(crv)

    # -------------------------------------------------------------------------
    #                               HK diagram
    # -------------------------------------------------------------------------

    def fit_linear_relation(self):
        C = self.linear_coefficients
        if C[0] == 0 and C[1] == 0 and C[2] == 0:
            K, H, k1, k2 = self.bspline.curvatures()
            c = fitting.linear_regression(np.column_stack((K, H)))
            n = c[0][0]
            b = np.sum(c[1] * n)
            return n[0], n[1], b
        else:
            return C

    # -------------------------------------------------------------------------
    #                               Initialization
    # -------------------------------------------------------------------------

    def set_dimensions(self):
        self._N1 = 3 * self.bspline.C
        self._N2 = self._N1 + 3 * self.bspline.S
        self._N3 = self._N2 + 4 * self.bspline.S
        self._N4 = self._N3 + 3 * self.bspline.S
        self._N5 = self._N4 + 2 * self.bspline.S
        N6 = self._N4 + 2 * self.bspline.S
        N = self._N4 + 2 * self.bspline.S
        if self.get_weight('isolines_parallel') > 0:
            N += 8 * self.bspline.S
            N6 += 8 * self.bspline.S
        if self._N != N:
            self.reinitialize()
        self._N6 = N6
        self._N = N

    def initialize_unknowns_vector(self):
        Base.initialize_unknowns_vector(self)
        X = self.bspline.control_points.flatten('F')
        Du, Dv = self.bspline.first_derivatives()
        N = np.cross(Du, Dv)
        X = np.hstack((X, N.flatten('F')))
        E, F, G, L, M, N, K, H = self.bspline.fundamental_forms()
        X = np.hstack((X, E, F, G, E * G - F ** 2, L, M, N, K, H))
        if self.get_weight('isolines_parallel') > 0:
            Ku, Kv = self.bspline.uv_central_differences(K)
            Hu, Hv = self.bspline.uv_central_differences(H)
            DK = (E * G - F ** 2) ** (-1) * (E * Kv ** 2 - 2 * F * Ku * Kv + G * Ku ** 2)
            DH = (E * G - F ** 2) ** (-1) * (E * Hv ** 2 - 2 * F * Hu * Hv + G * Hu ** 2)
            X = np.hstack((X, Ku ** 2, Ku * Kv, Kv ** 2, Hu ** 2, Hu * Hv, Hv ** 2, DK, DH))
        self._X = X
        self._X0 = np.copy(X)

    def make_errors(self):
        self.normals_error()
        self.first_error()
        self.determinant_error()
        self.second_error()
        self.gaussian_error()
        self.mean_error()
        self.squared_gradients_error()
        self.gradients_norms_error()
        self.hk_error()
        self.k1k2_error()

    def post_iteration_update(self):
        C = self.bspline.C
        self.bspline.control_points[:, 0] = self.X[0:C]
        self.bspline.control_points[:, 1] = self.X[C:2 * C]
        self.bspline.control_points[:, 2] = self.X[2 * C:3 * C]

    # -------------------------------------------------------------------------
    #                                Interaction
    # -------------------------------------------------------------------------

    def normalize_bspline(self):
        K, H, k1, k2 = self.bspline.curvatures()
        norm = np.mean(np.abs(np.hstack((k1, k2))))
        self.scale_bspline(norm * 5 * self._scale)

    def scale_bspline(self, factor):
        if self.normalize_geometry:
            self._bspline.scale(factor)
            for i in range(len(self.curves)):
                self.curves[i].control_points *= factor
        self.set_reference()
        self.reinitialize()
        c = self.bspline.curvatures()
        ref = (np.mean(np.abs(c[0])), np.mean(np.abs(c[1])))
        self.add_value('ref_curvatures', ref)
        self.__scale *= factor

    def rebuild_bspline(self, u_points=None, v_points=None,
                        u_degree=None, v_degree=None):
        self.bspline.rebuild(u_points, v_points, u_degree, v_degree)
        self.set_reference()
        self.reinitialize()

    def resize_bspline(self):
        self._bspline.scale(1/self.__scale)

    def unresize_bspline(self):
        self._bspline.scale(self.__scale)

    def reset_scaling(self):
        self.__scale = 1

    # -------------------------------------------------------------------------
    #                                  Results
    # -------------------------------------------------------------------------

    def control_points(self):
        C = self.bspline.C
        points = self.X[0:3 * C]
        points = np.reshape(points, (C, 3), order='F')
        return points

    def normals(self, initialized=False):
        if initialized:
            X = self._X0
        else:
            X = self._X
        N1 = self._N1
        S = self.bspline.S
        normals = X[N1:N1 + 3 * S]
        normals = np.reshape(normals, (S, 3), order='F')
        return normals

    def first(self, initialized=False):
        if initialized:
            X = self._X0
        else:
            X = self._X
        N2 = self._N2
        S = self.bspline.S
        EFG = X[N2:N2 + 3 * S]
        EFG = np.reshape(EFG, (S, 3), order='F')
        return EFG

    def determinant(self, initialized=False):
        if initialized:
            X = self._X0
        else:
            X = self._X
        N2 = self._N2
        S = self.bspline.S
        Det = X[N2 + 3 * S:N2 + 4 * S]
        return Det

    def second(self, initialized=False):
        if initialized:
            X = self._X0
        else:
            X = self._X
        N3 = self._N3
        S = self.bspline.S
        LMN = X[N3:N3 + 3 * S]
        LMN = np.reshape(LMN, (S, 3), order='F')
        return LMN

    def gaussian(self, initialized=False):
        if initialized:
            X = self._X0
        else:
            X = self._X
        N4 = self._N4
        S = self.bspline.S
        K = X[N4:N4 + S]
        return K

    def mean(self, initialized=False):
        if initialized:
            X = self._X0
        else:
            X = self._X
        N4 = self._N4
        S = self.bspline.S
        H = X[N4 + S:N4 + 2 * S]
        return H

    def squared_gradients(self, initialized=False):
        if self.get_weight('isolines_parallel') <= 0:
            return None
        if initialized:
            X = self._X0
        else:
            X = self._X
        N5 = self._N5
        S = self.bspline.S
        sg = X[N5:N5 + 6 * S]
        return sg.reshape((6, S), order='F')

    def gradients_norms(self, initialized=False):
        if self.get_weight('isolines_parallel') <= 0:
            return None
        if initialized:
            X = self._X0
        else:
            X = self._X
        N5 = self._N5
        S = self.bspline.S
        sg = X[N5 + 6 * S:N5 + 8 * S]
        return sg.reshape((S, 2), order='F')

    # -------------------------------------------------------------------------
    #                                  Errors
    # -------------------------------------------------------------------------

    def normals_error(self):
        N0 = self.normals(initialized=True)
        N = self.normals()
        e = np.linalg.norm(N - N0, axis=1) / np.median(np.linalg.norm(N0, axis=1))
        Emean = np.mean(e)
        Emax = np.max(e)
        self.add_error('normals', Emean, Emax, self.get_weight('normals'))

    def first_error(self):
        F0 = self.first(initialized=True)
        F = self.first()
        e = np.abs(np.sum(F - F0, axis=1) / (np.mean(np.abs(F0) + 1e-8)))
        Emean = np.mean(e)
        Emax = np.max(e)
        self.add_error('first', Emean, Emax, self.get_weight('first'))

    def determinant_error(self):
        D0 = self.determinant(initialized=True)
        D = self.determinant()
        e = np.abs((D - D0) / (np.mean(np.abs(D0) + 1e-8)))
        Emean = np.mean(e)
        Emax = np.max(e)
        self.add_error('determinant', Emean, Emax,
                       self.get_weight('determinant'))

    def second_error(self):
        S0 = self.second(initialized=True)
        S = self.second()
        e = np.abs(np.sum(S - S0, axis=1) / (np.mean(np.abs(S0) + 1e-8)))
        Emean = np.mean(e)
        Emax = np.max(e)
        self.add_error('second', Emean, Emax, self.get_weight('second'))

    def gaussian_error(self):
        K0 = self.gaussian(initialized=True)
        K = self.gaussian()
        e = np.abs((K - K0) / (np.mean(np.abs(K0) + 1e-8)))
        Emean = np.mean(e)
        Emax = np.max(e)
        self.add_error('gaussian', Emean, Emax, self.get_weight('gaussian'))

    def mean_error(self):
        H0 = self.mean(initialized=True)
        H = self.mean()
        e = np.abs((H - H0) / (np.mean(np.abs(H0) + 1e-8)))
        Emean = np.mean(e)
        Emax = np.max(e)
        self.add_error('mean', Emean, Emax, self.get_weight('mean'))

    def hk_error(self):
        K = self.gaussian(True)
        H = self.mean(True)
        p, up, um, vp, vm = self.bspline.uv_central_differences_iterators()
        e = ((K[up] - K[um]) * (H[vp] - H[vm]) - (K[vp] - K[vm]) * (H[up] - H[um]))
        Emean = np.mean(np.abs(e))
        Emax = np.max(np.abs(e))
        self.add_error('hk', Emean, Emax, self.get_weight('hk'))

    def k1k2_error(self):
        K = self.gaussian(True)
        H = self.mean(True)
        D = np.abs((H ** 2 - K)) ** .5
        p, up, um, vp, vm = self.bspline.uv_central_differences_iterators()
        M11 = H[up] + D[up] - H[um] - D[um]
        M21 = H[vp] + D[vp] - H[vm] - D[vm]
        M12 = H[up] - D[up] - H[um] + D[um]
        M22 = H[vp] - D[vp] - H[vm] + D[vm]
        e = np.abs(M11 * M22 - M12 * M21)
        Emean = np.mean(e)
        Emax = np.max(e)
        self.add_error('k1k2', Emean, Emax, self.get_weight('hk'))

    def squared_gradients_error(self):
        if self.get_weight('isolines_parallel') <= 0:
            return None
        SG = self.squared_gradients()
        SG0 = self.squared_gradients(True)
        e = np.abs(SG - SG0)
        n = 1 / (np.mean(np.abs(SG0), axis=1) + 1e-5)
        for i in range(6):
            e[:, i] *= n[i]
        Emean = np.mean(e[:, :3])
        Emax = np.max(e[:, :3])
        self.add_error('K_sq_gradients', Emean, Emax, self.get_weight(
            'squared_gradients'))
        Emean = np.mean(e[:, 3:])
        Emax = np.max(e[:, 3:])
        self.add_error('H_sq_gradients', Emean, Emax, self.get_weight(
            'squared_gradients'))

    def gradients_norms_error(self):
        if self.get_weight('isolines_parallel') <= 0:
            return None
        SG = self.gradients_norms()
        SG0 = self.gradients_norms(True)
        e = np.abs(SG - SG0)
        n = 1 / (np.mean(np.abs(SG0), axis=0) + 1e-5)
        for i in range(2):
            e[:, i] *= n[i]
        Emean = np.mean(e[:, 0])
        Emax = np.max(e[:, 0])
        self.add_error('K_grad_norms', Emean, Emax, self.get_weight(
            'gradients_norms'))
        Emean = np.mean(e[:, 1])
        Emax = np.max(e[:, 1])
        self.add_error('H_grad_norms', Emean, Emax, self.get_weight(
            'gradients_norms'))

    def geometric_error(self):
        e_mean = 0
        e_max = 0
        n = 0
        errors = ['normals', 'first', 'second', 'determinant',
                  'gaussian', 'mean']
        for key in errors:
            err = self.get_error(key)
            e_mean += err[0]
            e_max = max(err[1], e_max)
            n += 1
        return e_mean / n, e_max

    # -------------------------------------------------------------------------
    #                                   Utilities
    # -------------------------------------------------------------------------

    def control_lengths(self):
        v1, v2 = self._bspline.edge_vertices()
        E = self._bspline.control_points[v1] - self._bspline.control_points[v2]
        l = np.linalg.norm(E, axis=1)
        return l

    def control_diagonals(self):
        v1, v2 = self._bspline.face_diagonals()
        E = self._bspline.control_points[v1] - self._bspline.control_points[v2]
        l = np.linalg.norm(E, axis=1)
        return l

    def info_string(self):
        out = str(self.bspline) + '\n'
        out += 'sampling: ({},{})'.format(self.bspline.u_sampling,
                                          self.bspline.v_sampling)
        out += 'reference: {:.4f}\n'.format(self.get_weight('reference'))
        out += 'control fairness: {:.4f}\n'.format(self.get_weight(
            'control_fairness'))
        out += 'isolines fairness: {:.4f}\n'.format(self.get_weight(
            'curvature_fairness'))

    def principal_curvatures(self, initialized=True):
        H = self.mean(initialized)
        K = self.gaussian(initialized)
        d = np.abs(H ** 2 - K) ** .5
        k1 = H + d
        k2 = H - d
        return k1, k2

    def plot_surface(self):
        plot = []
        self.bspline.plot_mode = 'control'
        plot.append(plotter.Edges(self.bspline))
        plot.append(plotter.Points(self.bspline.control_points,
                                       radius=0.15, color='gold'))
        self.bspline.plot_mode = 'surface'
        plot.append(plotter.Faces(self.bspline))
        plotter.view(plot)

    # --------------------------------------------------------------------------
    #                                Errors strings
    # --------------------------------------------------------------------------

    def geometric_error_string(self):
        return '{:.4E} | {:.4E}'.format(*self.geometric_error())

    # -------------------------------------------------------------------------
    #                                  Constraints
    # -------------------------------------------------------------------------

    def forms_constraints(self):
        C = self.bspline.C
        S = self.bspline.S
        N = self.N
        X = self._X
        s, u, v, c, cu, cv = self.bspline.uv_sampling_iterators()
        Nu, Nv = self.bspline.basis_functions()
        Nuu, Nvv = self.bspline.derivative_basis_functions(d=1)
        Nuuu, Nvvv = self.bspline.derivative_basis_functions(d=2)
        Du, Dv, Duu, Duv, Dvv = self.bspline.second_derivatives()
        s0 = np.arange(S)
        one = np.ones(S)
        # ----------------------------------------------------------------------
        #                             normals
        # ----------------------------------------------------------------------
        w = self.get_weight('normals')
        i = np.hstack((s, s, s, s, s0))
        # ----------------------------------------------------------------------
        nx0 = self._N1 + s0
        j = np.hstack((C + c, 2 * C + c, C + c, 2 * C + c, nx0))
        data = np.hstack((Nuu[u, cu] * Nv[v, cv] * Dv[s, 2],
                          Nu[u, cu] * Nvv[v, cv] * Du[s, 1],
                          -Nu[u, cu] * Nvv[v, cv] * Du[s, 2],
                          -Nuu[u, cu] * Nv[v, cv] * Dv[s, 1],
                          -one)) * w
        r = (Du[s0, 1] * Dv[s0, 2] - Du[s0, 2] * Dv[s0, 1]) * w
        H = sparse.coo_matrix((data, (i, j)), shape=(S, N))
        self.add_iterative_constraint(H, r, 'Nx')
        # ----------------------------------------------------------------------
        ny0 = nx0 + S
        j = np.hstack((2 * C + c, c, 2 * C + c, c, ny0))
        data = np.hstack((Nuu[u, cu] * Nv[v, cv] * Dv[s, 0],
                          Nu[u, cu] * Nvv[v, cv] * Du[s, 2],
                          -Nu[u, cu] * Nvv[v, cv] * Du[s, 0],
                          -Nuu[u, cu] * Nv[v, cv] * Dv[s, 2],
                          -one)) * w
        r = (Du[s0, 2] * Dv[s0, 0] - Du[s0, 0] * Dv[s0, 2]) * w
        H = sparse.coo_matrix((data, (i, j)), shape=(S, N))
        self.add_iterative_constraint(H, r, 'Ny')
        # ----------------------------------------------------------------------
        nz0 = ny0 + S
        j = np.hstack((c, C + c, c, C + c, nz0))
        data = np.hstack((Nuu[u, cu] * Nv[v, cv] * Dv[s, 1],
                          Nu[u, cu] * Nvv[v, cv] * Du[s, 0],
                          -Nu[u, cu] * Nvv[v, cv] * Du[s, 1],
                          -Nuu[u, cu] * Nv[v, cv] * Dv[s, 0],
                          -one)) * w
        r = (Du[s0, 0] * Dv[s0, 1] - Du[s0, 1] * Dv[s0, 0]) * w
        H = sparse.coo_matrix((data, (i, j)), shape=(S, N))
        self.add_iterative_constraint(H, r, 'Nz')
        # ----------------------------------------------------------------------
        #                       first fundamental form
        # ----------------------------------------------------------------------
        w = self.get_weight('first')
        # ----------------------------------------------------------------------
        e0 = self._N2 + s0
        i = np.hstack((s, s, s, s0))
        j = np.hstack((c, C + c, 2 * C + c, e0))
        data = np.hstack((2 * Nuu[u, cu] * Nv[v, cv] * Du[s, 0],
                          2 * Nuu[u, cu] * Nv[v, cv] * Du[s, 1],
                          2 * Nuu[u, cu] * Nv[v, cv] * Du[s, 2], -one)) * w
        r = np.einsum('ij,ij->i', Du[s0], Du[s0]) * w
        H = sparse.coo_matrix((data, (i, j)), shape=(S, N))
        self.add_iterative_constraint(H, r, 'E')
        # ----------------------------------------------------------------------
        f0 = e0 + S
        i = np.hstack((s, s, s, s, s, s, s0))
        j = np.hstack((c, C + c, 2 * C + c, c, C + c, 2 * C + c, f0))
        data = np.hstack((Nuu[u, cu] * Nv[v, cv] * Dv[s, 0],
                          Nuu[u, cu] * Nv[v, cv] * Dv[s, 1],
                          Nuu[u, cu] * Nv[v, cv] * Dv[s, 2],
                          Nu[u, cu] * Nvv[v, cv] * Du[s, 0],
                          Nu[u, cu] * Nvv[v, cv] * Du[s, 1],
                          Nu[u, cu] * Nvv[v, cv] * Du[s, 2],
                          -one)) * w
        r = np.einsum('ij,ij->i', Du[s0], Dv[s0]) * w
        H = sparse.coo_matrix((data, (i, j)), shape=(S, N))
        self.add_iterative_constraint(H, r, 'F')
        # ----------------------------------------------------------------------
        g0 = f0 + S
        i = np.hstack((s, s, s, s0))
        j = np.hstack((c, C + c, 2 * C + c, g0))
        data = np.hstack((2 * Nu[u, cu] * Nvv[v, cv] * Dv[s, 0],
                          2 * Nu[u, cu] * Nvv[v, cv] * Dv[s, 1],
                          2 * Nu[u, cu] * Nvv[v, cv] * Dv[s, 2], -one)) * w
        r = np.einsum('ij,ij->i', Dv[s0], Dv[s0]) * w
        H = sparse.coo_matrix((data, (i, j)), shape=(S, N))
        self.add_iterative_constraint(H, r, 'G')
        # ---------------------------------------------------------------------
        #                       second fundamental form
        # ---------------------------------------------------------------------
        w = self.get_weight('second')
        nx = self._N1 + s
        ny = nx + S
        nz = ny + S
        Nn = self.normals()
        det = np.abs(self.determinant()) ** .5
        i = np.hstack((s, s, s, s0, s0, s0, s0))
        # ---------------------------------------------------------------------
        l = self._N3 + s0
        j = np.hstack((c, C + c, 2 * C + c, nx0, ny0, nz0, l))
        data = np.hstack((Nuuu[u, cu] * Nv[v, cv] * X[nx] / det[s],
                          Nuuu[u, cu] * Nv[v, cv] * X[ny] / det[s],
                          Nuuu[u, cu] * Nv[v, cv] * X[nz] / det[s],
                          Duu[s0, 0] / det[s0], Duu[s0, 1] / det[s0],
                          Duu[s0, 2] / det[s0], -one)) * w
        r = np.einsum('ij,ij,i->i', Duu[s0], Nn[s0], 1 / det[s0]) * w
        H = sparse.coo_matrix((data, (i, j)), shape=(S, N))
        self.add_iterative_constraint(H, r, 'L')
        # ---------------------------------------------------------------------
        m = l + S
        j = np.hstack((c, C + c, 2 * C + c, nx0, ny0, nz0, m))
        data = np.hstack((Nuu[u, cu] * Nvv[v, cv] * X[nx] / det[s],
                          Nuu[u, cu] * Nvv[v, cv] * X[ny] / det[s],
                          Nuu[u, cu] * Nvv[v, cv] * X[nz] / det[s],
                          Duv[s0, 0] / det[s0], Duv[s0, 1] / det[s0],
                          Duv[s0, 2] / det[s0], -one)) * w
        r = np.einsum('ij,ij,i->i', Duv[s0], Nn[s0], 1 / det[s0]) * w
        H = sparse.coo_matrix((data, (i, j)), shape=(S, N))
        self.add_iterative_constraint(H, r, 'M')
        # ---------------------------------------------------------------------
        n = m + S
        j = np.hstack((c, C + c, 2 * C + c, nx0, ny0, nz0, n))
        data = np.hstack((Nu[u, cu] * Nvvv[v, cv] * X[nx] / det[s],
                          Nu[u, cu] * Nvvv[v, cv] * X[ny] / det[s],
                          Nu[u, cu] * Nvvv[v, cv] * X[nz] / det[s],
                          Dvv[s0, 0] / det[s0], Dvv[s0, 1] / det[s0],
                          Dvv[s0, 2] / det[s0], -one)) * w
        r = np.einsum('ij,ij,i->i', Dvv[s0], Nn[s0], 1 / det[s0]) * w
        H = sparse.coo_matrix((data, (i, j)), shape=(S, N))
        self.add_iterative_constraint(H, r, 'N')
        # ---------------------------------------------------------------------

    def determinant_constraints(self):
        w = self.get_weight('determinant')
        S = self.bspline.S
        N = self.N
        X = self._X
        s = np.arange(S)
        e = self._N2 + s
        f = e + S
        g = f + S
        det = g + S
        i = np.hstack((s, s, s, s))
        j = np.hstack((e, g, f, det))
        data = np.hstack((X[g], X[e], -2 * X[f], -np.ones(S))) * w
        r = (X[e] * X[g] - X[f] ** 2) * w
        H = sparse.coo_matrix((data, (i, j)), shape=(S, N))
        self.add_iterative_constraint(H, r, 'Det')

    def curvature_constraints(self):
        S = self.bspline.S
        N = self.N
        X = self._X
        s = np.arange(S)
        e = self._N2 + s
        f = e + S
        g = f + S
        det = g + S
        l = self._N3 + s
        m = l + S
        n = m + S
        # ---------------------------------------------------------------------
        w = self.get_weight('gaussian')
        k = self._N4 + s
        i = np.hstack((s, s, s, s, s))
        j = np.hstack((l, m, n, k, det))
        data = np.hstack((X[n], -2 * X[m], X[l], -X[det], -X[k])) * w
        r = (X[n] * X[l] - X[m] ** 2 - X[k] * X[det]) * w
        H = sparse.coo_matrix((data, (i, j)), shape=(S, N))
        self.add_iterative_constraint(H, r, 'K')
        # ---------------------------------------------------------------------
        # H = (LG - 2MF + NE) / 2(EG - F^2)
        w = self.get_weight('mean')
        h = self._N4 + S + s
        i = np.hstack((s, s, s, s, s, s, s, s))
        j = np.hstack((e, f, g, l, m, n, h, det))
        data = np.hstack((X[n], -2 * X[m], X[l], X[g], -2 * X[f], X[e],
                          -2 * X[det], -2 * X[h])) * w
        r = (X[l] * X[g] - 2 * X[m] * X[f] + X[n] * X[e] - 2 * X[h] * X[det]) * w
        H = sparse.coo_matrix((data, (i, j)), shape=(S, N))
        self.add_iterative_constraint(H, r, 'H')

    def hk_constraints(self):
        w = self.get_weight('hk')
        N = self.N
        p, up, um, vp, vm = self.bspline.uv_central_differences_iterators(
            include_boundary=self.include_boundary)
        Hu, Hv = self.bspline.uv_central_differences(self.mean())
        Ku, Kv = self.bspline.uv_central_differences(self.gaussian())
        Hu = Hu[p]
        Hv = Hv[p]
        Ku = Ku[p]
        Kv = Kv[p]
        k = self._N4
        h = self._N4 + self.bspline.S
        S = len(p)
        Kn = (Ku ** 2 + Kv ** 2) ** .5 + 1e-8
        Hn = (Hu ** 2 + Hv ** 2) ** .5 + 1e-8
        nr = np.mean(Kn * Hn) / (Kn * Hn) * 100
        nr = np.minimum(nr, 10 * np.mean(nr))
        # nr = 100
        # nr = np.mean(Kn) / Kn * 100
        i = np.arange(S)
        i = np.hstack((i, i, i, i, i, i, i, i))
        j = np.hstack((k + up, k + um, h + vp, h + vm, k + vp, k + vm,
                       h + up, h + um))
        data = np.hstack((Hv * nr, -Hv * nr, Ku * nr, -Ku * nr,
                          -Hu * nr, Hu * nr, -Kv * nr, Kv * nr)) * w
        r = (Ku * Hv - Kv * Hu) * nr * w
        H = sparse.coo_matrix((data, (i, j)), shape=(S, N))
        self.add_iterative_constraint(H, r, 'Wein')

    def hk_constraints2(self):
        w = self.get_weight('hk')
        S = self.bspline.S
        N = self.N
        X = self._X
        p, up, um, vp, vm = self.bspline.uv_central_differences_iterators()
        K = self._N4
        H = self._N4 + S
        i = np.arange(S)
        i = np.hstack((i, i, i, i, i, i, i, i))
        j = np.hstack((K + up, K + um, H + vp, H + vm, K + vp, K + vm, H + up, H + um))
        data = np.hstack(((X[H + vp] - X[H + vm]), -(X[H + vp] - X[H + vm]),
                          (X[K + up] - X[K + um]), -(X[K + up] - X[K + um]),
                          -(X[H + up] - X[H + um]), (X[H + up] - X[H + um]),
                          -(X[K + vp] - X[K + vm]), (X[K + vp] - X[K + vm]))) * w
        r = ((X[K + up] - X[K + um]) * (X[H + vp] - X[H + vm]) -
             (X[K + vp] - X[K + vm]) * (X[H + up] - X[H + um])) * w
        H = sparse.coo_matrix((data, (i, j)), shape=(S, N))
        self.add_iterative_constraint(H, r, 'Wein')

    # -------------------------------------------------------------------------
    #                             Isolines Fairness
    # -------------------------------------------------------------------------

    def squared_gradients_constraints(self):
        if self.get_weight('isolines_parallel') <= 0:
            return
        S = self.bspline.S
        k = self._N4
        h = self._N4 + S
        p, up, um, vp, vm = self.bspline.uv_central_differences_iterators()
        Hu, Hv, du, dv = self.bspline.uv_central_differences(self.mean(), True)
        Ku, Kv = self.bspline.uv_central_differences(self.gaussian())
        s = np.arange(S)
        one = np.ones(S)
        ku2 = self._N5 + s
        kukv = self._N5 + S + s
        kv2 = self._N5 + 2 * S + s
        hu2 = self._N5 + 3 * S + s
        huhv = self._N5 + 4 * S + s
        hv2 = self._N5 + 5 * S + s
        # ---------------------------------------------------------------------
        w = self.get_weight('squared_gradients')
        i1 = np.hstack((s, s, s))
        i2 = np.hstack((s, s, s, s, s))
        # ---------------------------------------------------------------------
        j = np.hstack((k + up, k + um, ku2))
        data = np.hstack((2 * Ku / du, - 2 * Ku / du, -one)) * w
        H = sparse.coo_matrix((data, (i1, j)), shape=(S, self.N))
        r = Ku ** 2 * w
        self.add_iterative_constraint(H, r, 'Ku^2')
        # ---------------------------------------------------------------------
        j = np.hstack((k + vp, k + vm, kv2))
        data = np.hstack((2 * Kv / dv, - 2 * Kv / dv, -one)) * w
        H = sparse.coo_matrix((data, (i1, j)), shape=(S, self.N))
        r = Kv ** 2 * w
        self.add_iterative_constraint(H, r, 'Kv^2')
        # ---------------------------------------------------------------------
        j = np.hstack((k + up, k + um, k + vp, k + vm, kukv))
        data = np.hstack((Kv / du, -Kv / du, Ku / dv, -Ku / dv, -one)) * w
        H = sparse.coo_matrix((data, (i2, j)), shape=(S, self.N))
        r = Ku * Kv * w
        self.add_iterative_constraint(H, r, 'KuKv')
        # ---------------------------------------------------------------------
        w = self.get_weight('squared_gradients')
        j = np.hstack((h + up, h + um, hu2))
        data = np.hstack((2 * Hu / du, - 2 * Hu / du, -one)) * w
        H = sparse.coo_matrix((data, (i1, j)), shape=(S, self.N))
        r = Hu ** 2 * w
        self.add_iterative_constraint(H, r, 'Hu^2')
        # ---------------------------------------------------------------------
        j = np.hstack((h + vp, h + vm, hv2))
        data = np.hstack((2 * Hv / dv, - 2 * Hv / dv, -one)) * w
        H = sparse.coo_matrix((data, (i1, j)), shape=(S, self.N))
        r = Hv ** 2 * w
        self.add_iterative_constraint(H, r, 'Hv^2')
        # ---------------------------------------------------------------------
        j = np.hstack((h + up, h + um, h + vp, h + vm, huhv))
        data = np.hstack((Hv / du, -Hv / du, Hu / dv, -Hu / dv, -one)) * w
        H = sparse.coo_matrix((data, (i2, j)), shape=(S, self.N))
        r = Hu * Hv * w
        self.add_iterative_constraint(H, r, 'HuHv')

    def gradients_norm_constraints(self):
        if self.get_weight('isolines_parallel') <= 0:
            return
        S = self.bspline.S
        s = np.arange(S)
        one = np.ones(S)
        X = self._X
        e = self._N2 + s
        f = e + S
        g = f + S
        det = X[g + S]
        ku2 = self._N5 + s
        kukv = self._N5 + S + s
        kv2 = self._N5 + 2 * S + s
        hu2 = self._N5 + 3 * S + s
        huhv = self._N5 + 4 * S + s
        hv2 = self._N5 + 5 * S + s
        dk = self._N5 + 6 * S + s
        dh = self._N5 + 7 * S + s
        # ---------------------------------------------------------------------
        # DK = (E*G-F**2)**(-1) * (E*Kv**2 - 2*F*Ku*Kv + G*Ku**2)
        # DH = (E*G-F**2)**(-1) * (E*Hv**2 - 2*F*Hu*Hv + G*Hu**2)
        # ---------------------------------------------------------------------
        i = np.hstack((s, s, s, s, s, s, s))
        # ---------------------------------------------------------------------
        w = self.get_weight('gradients_norms')
        j = np.hstack((e, f, g, kv2, kukv, ku2, dk))
        data = np.hstack((X[kv2] / det, -2 * X[kukv] / det, X[ku2] / det,
                          X[e] / det, -2 * X[f] / det, X[g] / det, -one)) * w
        H = sparse.coo_matrix((data, (i, j)), shape=(S, self.N))
        r = (X[e] * X[kv2] - 2 * X[f] * X[kukv] + X[g] * X[ku2]) / det * w
        self.add_iterative_constraint(H, r, 'DK')
        # ---------------------------------------------------------------------
        w = self.get_weight('gradients_norms')
        j = np.hstack((e, f, g, hv2, huhv, hu2, dh))
        data = np.hstack((X[hv2] / det, -2 * X[huhv] / det, X[hu2] / det,
                          X[e] / det, -2 * X[f] / det, X[g] / det, -one)) * w
        H = sparse.coo_matrix((data, (i, j)), shape=(S, self.N))
        r = (X[e] * X[hv2] - 2 * X[f] * X[huhv] + X[g] * X[hu2]) / det * w
        self.add_iterative_constraint(H, r, 'DH')

    def isolines_parallel_constraints2(self):
        if self.get_weight('isolines_parallel') <= 0:
            return
        S = self.bspline.S
        N = self.N
        X = self._X
        p, up, um, vp, vm = self.bspline.uv_central_differences_iterators()
        K = self._N4
        dK = self._N5 + 6 * S
        i = np.arange(S)
        w = self.get_weight('isolines_parallel') * 100
        i = np.hstack((i, i, i, i, i, i, i, i))
        j = np.hstack((K + up, K + um, dK + vp, dK + vm, K + vp, K + vm,
                       dK + up, dK + um))
        data = np.hstack(((X[dK + vp] - X[dK + vm]), -(X[dK + vp] - X[dK + vm]),
                          (X[K + up] - X[K + um]), -(X[K + up] - X[K + um]),
                          -(X[dK + up] - X[dK + um]), (X[dK + up] - X[dK + um]),
                          -(X[K + vp] - X[K + vm]), (X[K + vp] - X[K + vm]))) * w
        r = ((X[K + up] - X[K + um]) * (X[dK + vp] - X[dK + vm]) -
             (X[K + vp] - X[K + vm]) * (X[dK + up] - X[dK + um])) * w
        H = sparse.coo_matrix((data, (i, j)), shape=(S, N))
        self.add_iterative_constraint(H, r, 'Kfair')
        w = self.get_weight('isolines_parallel')
        H = self._N4 + S
        dH = self._N5 + 7 * S
        i = np.arange(S)
        i = np.hstack((i, i, i, i, i, i, i, i))
        # print(np.mean(np.abs(X[dK+vp]-X[dK+vm])))
        # print(np.mean(np.abs(X[dH+vp]-X[dH+vm])))
        j = np.hstack((H + up, H + um, dH + vp, dH + vm, H + vp, H + vm, dH + up, dH + um))
        data = np.hstack(((X[dH + vp] - X[dH + vm]), -(X[dK + vp] - X[dH + vm]),
                          (X[H + up] - X[H + um]), -(X[H + up] - X[H + um]),
                          -(X[dH + up] - X[dH + um]), (X[dH + up] - X[dH + um]),
                          -(X[H + vp] - X[H + vm]), (X[H + vp] - X[H + vm]))) * w
        r = ((X[H + up] - X[H + um]) * (X[dH + vp] - X[dH + vm]) -
             (X[H + vp] - X[H + vm]) * (X[dH + up] - X[dH + um])) * w
        H = sparse.coo_matrix((data, (i, j)), shape=(S, N))
        self.add_iterative_constraint(H, r, 'Hfair')

    def isolines_parallel_constraints(self):
        w = self.get_weight('isolines_parallel')
        if w <= 0:
            return
        N = self.N
        X = self._X
        norms = self.gradients_norms()
        p, up, um, vp, vm = self.bspline.uv_central_differences_iterators(
            include_boundary=False)
        S = len(p)
        i = np.arange(S)
        i = np.hstack((i, i, i, i, i, i, i, i))
        # ---------------------------------------------------------------------
        K = self._N4
        dK = self._N5 + 6 * self.bspline.S
        Ku, Kv = self.bspline.uv_central_differences(self.gaussian())
        dKu, dKv = self.bspline.uv_central_differences(norms[:, 0])
        Ku = Ku[p]
        Kv = Kv[p]
        dKu = dKu[p]
        dKv = dKv[p]
        Kn = (Ku ** 2 + Kv ** 2) ** .5 + 1e-8
        dKn = (dKu ** 2 + dKv ** 2) ** .5 + 1e-8
        nr = np.mean(Kn * dKn) / (Kn * dKn) * 1
        #nr = 0.1
        j = np.hstack((K + up, K + um, dK + vp, dK + vm, K + vp, K + vm,
                       dK + up, dK + um))
        data = np.hstack((dKv * nr, -dKv * nr, Ku * nr, -Ku * nr,
                          -dKu * nr, dKu * nr, -Kv * nr, Kv * nr)) * w
        r = (Ku * dKv - Kv * dKu) * nr * w
        H = sparse.coo_matrix((data, (i, j)), shape=(S, N))
        self.add_iterative_constraint(H, r, 'KNfair')
        # ---------------------------------------------------------------------
        H = self._N4 + S
        dH = self._N5 + 7 * S
        Hu, Hv = self.bspline.uv_central_differences(self.mean())
        dHu, dHv = self.bspline.uv_central_differences(norms[:, 1])
        Hu = Hu[p]
        Hv = Hv[p]
        dHu = dHu[p]
        dHv = dHv[p]
        Hn = (Hu ** 2 + Hv ** 2) ** .5 + 1e-8
        dHn = (dHu ** 2 + dHv ** 2) ** .5 + 1e-8
        nr = np.mean(Hn * dHn) / (Hn * dHn) * 10
        #nr = 1
        j = np.hstack((H + up, H + um, dH + vp, dH + vm, H + vp, H + vm,
                       dH + up, dH + um))
        data = np.hstack((dHv * nr, -dHv * nr, Hu * nr, -Hu * nr,
                          -dHu * nr, dHu * nr, -Hv * nr, Hv * nr)) * w
        r = (Hu * dHv - Hv * dHu) * nr * w
        H = sparse.coo_matrix((data, (i, j)), shape=(S, N))
        self.add_iterative_constraint(H, r, 'HNfair')

    # -------------------------------------------------------------------------
    #                              Linear Weingarten
    # -------------------------------------------------------------------------

    def linear_fitting_constraints(self):
        w = self.get_weight('linear_fitting')
        if w <= 0:
            return
        S = self.bspline.S
        s = np.arange(S)
        k = self._N4 + s
        h = self._N4 + S + s
        C = self.fit_linear_relation()
        a = np.repeat(C[0], S)
        b = np.repeat(C[1], S)
        c = np.repeat(C[2], S)
        i = np.hstack((s, s))
        j = np.hstack((k, h))
        data = np.hstack((a, b)) * w
        r = c * w
        H = sparse.coo_matrix((data, (i, j)), shape=(S, self.N))
        self.add_iterative_constraint(H, r, 'Lin')

    def curve_fitting_constraints0(self):
        w = self.get_weight('curve_fitting')
        if w <= 0:
            return
        for l in range(len(self._curves)):
            S = self.bspline.S
            s = np.arange(S)
            K = self.gaussian()
            H = self.mean()
            P = np.column_stack((K, H, np.zeros(len(K))))
            if self.adaptive_curve:
                n = self.n_curve_control_points
                try:
                    self._curve.fit_line_to_points(P, control_points_number=n)
                    self._curve.fit_to_points(P, straightening=0.1)
                except:
                    self.fit_curve()
            k = self._N4 + s
            h = self._N4 + S + s
            c = self._curve.closest_points(P)
            T, N = self._curve.frame(c)
            C = self._curve.points(c)[:, [0, 1]]
            N = N[:, [0, 1]]
            i = np.hstack((s, s))
            j = np.hstack((k, h))
            data = N.flatten('F') * w
            H = sparse.coo_matrix((data, (i, j)), shape=(S, self.N))
            r = np.einsum('ij,ij->i', N, C) * w
            self.add_iterative_constraint(H, r, 'Crv_t')
            i = np.hstack((s, S + s))
            data = np.ones(2 * S) * 0.01 * w
            H = sparse.coo_matrix((data, (i, j)), shape=(2 * S, self.N))
            r = C.flatten('F') * 0.01 * w
            self.add_iterative_constraint(H, r, 'Crv_{}'.format(l))

    # -------------------------------------------------------------------------
    #                                 Reference
    # -------------------------------------------------------------------------

    def reference_constraints(self):
        w = self.get_weight('reference') / 10
        C = 3 * self.bspline.C
        i = np.arange(C)
        H = sparse.coo_matrix((np.ones(C) * w, (i, i)), shape=(C, self.N))
        r = self.get_value('reference').flatten('F') * w
        self.add_constant_constraint(H, r, 'Ref')

    def fixed_points_constraints(self):
        w = self.get_weight('fixed')
        fix = np.unique(np.hstack((self.handle, self._fixed_points)))
        if len(fix) == 0:
            return
        C = 3 * len(fix)
        i = np.arange(C)
        j = np.hstack((fix, fix + self.bspline.C, fix + 2 * self.bspline.C))
        H = sparse.coo_matrix((np.ones(C) * w, (i, j)), shape=(C, self.N))
        r = self.bspline.control_points[fix].flatten('F') * w
        self.add_constant_constraint(H, r, 'Fix')

    def gliding_boundary_constraints(self):
        w = self.get_weight('gliding')
        if w <= 0:
            return
        bdrs = self.bspline.boundary_control_curves()
        for b in range(len(bdrs)):
            bdr = bdrs[b]
            crv = self._boundary_curves[b]
            c = crv.closest_vertices(self.bspline.control_points[bdr])
            T = crv.vertex_tangents(normalized=True)[c]
            V1 = utilities.orthogonal_vectors(T)
            V2 = np.cross(T, V1)
            C = len(c)
            nC = self.bspline.C
            i = np.arange(C)
            i = np.hstack((i, i, i))
            j = np.hstack((bdr, bdr + nC, bdr + 2 * nC))
            data = V1.flatten('F') * w
            r = np.einsum('ij,ij->i', self.bspline.control_points[bdr], V1) * w
            H = sparse.coo_matrix((data, (i, j)), shape=(C, self.N))
            self.add_iterative_constraint(H, r, 'Glid_{}_1'.format(b))
            data = V2.flatten('F') * w
            r = np.einsum('ij,ij->i', self.bspline.control_points[bdr], V2) * w
            H = sparse.coo_matrix((data, (i, j)), shape=(C, self.N))
            self.add_iterative_constraint(H, r, 'Glid_{}_2'.format(b))

    def control_lengths_constraints(self):
        w = self.get_weight('control_lengths')
        if w <= 0:
            return
        L = self.get_value('control_lengths')
        Lx = self.control_lengths()
        X = self.X
        v1, v2 = self.bspline.edge_vertices()
        E = len(v1)
        C = self.bspline.C
        v1 = np.hstack((v1, C + v1, 2 * C + v1))
        v2 = np.hstack((v2, C + v2, 2 * C + v2))
        e = np.arange(E)
        i = np.hstack((e, e, e, e, e, e))
        j = np.hstack((v1, v2))
        data = 2 * np.hstack((X[v1] - X[v2], X[v2] - X[v1])) * w
        H = sparse.coo_matrix((data, (i, j)), shape=(E, self.N))
        r = (Lx ** 2 + L ** 2) * w
        self.add_iterative_constraint(H, r, 'Len')

    def control_diagonals_constraints(self):
        w = self.get_weight('control_lengths') / 2
        if w <= 0:
            return
        L = self.get_value('control_diagonals')
        Lx = self.control_diagonals()
        X = self.X
        v1, v2 = self.bspline.face_diagonals()
        E = len(v1)
        C = self.bspline.C
        v1 = np.hstack((v1, C + v1, 2 * C + v1))
        v2 = np.hstack((v2, C + v2, 2 * C + v2))
        e = np.arange(E)
        i = np.hstack((e, e, e, e, e, e))
        j = np.hstack((v1, v2))
        data = 2 * np.hstack((X[v1] - X[v2], X[v2] - X[v1])) * w
        H = sparse.coo_matrix((data, (i, j)), shape=(E, self.N))
        r = (Lx ** 2 + L ** 2) * w
        self.add_iterative_constraint(H, r, 'Diag')

    # -------------------------------------------------------------------------
    #                                 Fairness
    # -------------------------------------------------------------------------

    def control_fairness_constraints(self):
        f = self.bspline.S ** .5 / 1000
        w = self.get_weight('control_fairness') * f
        c, cp, cm = self.bspline.control_points_curves_iterators()
        L = len(c)
        one = self.bspline.control_points[cp] - self.bspline.control_points[cm]
        one = np.linalg.norm(one, axis=1) ** -1
        C = self.bspline.C
        # one = np.ones(L)
        i = np.arange(L)
        i = np.hstack((i, i, i))
        j = np.hstack((c, cp, cm))
        data = np.hstack((-2 * one, one, one)) * w
        i = np.hstack((i, L + i, 2 * L + i))
        j = np.hstack((j, C + j, 2 * C + j))
        data = np.hstack((data, data, data))
        K = sparse.coo_matrix((data, (i, j)), shape=(3 * L, self.N))
        s = np.zeros((3 * L))
        self.add_iterative_fairness(K, s, 'CFair')

    def curvature_fairness_constraints(self):
        c = self.get_value('ref_curvatures')
        p, up, um, vp, vm = self.bspline.uv_central_differences_iterators(
            include_boundary=False)
        S = len(p)
        k = self._N4
        h = self._N4 + self.bspline.S
        one = np.ones(S)
        i = np.hstack((p, p, p, p, p))
        s = np.zeros(self.bspline.S)
        data = np.hstack((-one, .25 * one, .25 * one, .25 * one, .25 * one))
        # ---------------------------------------------------------------------
        w = self.get_weight('curvature_fairness') * c[1] / c[0]
        j = np.hstack((k + p, k + up, k + um, k + vp, k + vm))
        K = sparse.coo_matrix((data * w, (i, j)), shape=(self.bspline.S, self.N))
        self.add_iterative_fairness(K, s, 'KFair')
        # ---------------------------------------------------------------------
        w = self.get_weight('curvature_fairness')
        j = np.hstack((h + p, h + up, h + um, h + vp, h + vm))
        K = sparse.coo_matrix((data * w, (i, j)), shape=(self.bspline.S, self.N))
        self.add_iterative_fairness(K, s, 'HFair')

    def curvature_minimization_constraints(self):
        c = self.get_value('ref_curvatures')
        S = self.bspline.S
        i = np.arange(S)
        k = self._N4 + i
        h = self._N4 + S + i
        w = self.get_weight('curvature_minimization') * c[1] / c[0]
        data = np.ones(S)
        K = sparse.coo_matrix((data * w, (i, k)), shape=(S, self.N))
        s = np.zeros(S)
        self.add_iterative_fairness(K, s, 'Kmin')
        w = self.get_weight('curvature_minimization')
        K = sparse.coo_matrix((data * w, (i, h)), shape=(S, self.N))
        self.add_iterative_fairness(K, s, 'Hmin')

    # -------------------------------------------------------------------------
    #                                 Build
    # -------------------------------------------------------------------------

    def build_iterative_constraints(self):
        self.forms_constraints()
        self.determinant_constraints()
        self.curvature_constraints()
        self.hk_constraints()
        self.squared_gradients_constraints()
        self.gradients_norm_constraints()
        self.isolines_parallel_constraints()
        self.control_lengths_constraints()
        self.control_diagonals_constraints()
        self.linear_fitting_constraints()
        self.gliding_boundary_constraints()

    def build_constant_constraints(self):
        self.reference_constraints()
        self.fixed_points_constraints()

    def build_constant_fairness(self):
        pass

    def build_iterative_fairness(self):
        self.control_fairness_constraints()
        self.curvature_fairness_constraints()






#from geolab.optimization.sparse_dot_mkl.sparse_dot import (dot_product_mkl, dot_product_transpose_mkl, get_version_string, gram_matrix_mkl,
                                                #sparse_qr_solve_mkl, set_debug_mode)



# -*- coding: utf-8 -*-
from __future__ import absolute_import

from __future__ import print_function

from __future__ import division

import numpy as np

from scipy.sparse import coo_matrix

# -----------------------------------------------------------------------------

from geolab import utilities

from geolab.optimization import spsolve

# -----------------------------------------------------------------------------

__author__ = 'Davide Pellis'

# -----------------------------------------------------------------------------

def quadratic_fitting(mesh, double_ring=False, interpolate_vertex=False):
    e3 = mesh.vertex_normals()
    e1 = utilities.orthogonal_vectors(e3)
    e2 = np.cross(e3,e1)
    if double_ring:
        vi, vj, = mesh.vertex_double_ring_vertices_iterators()
    else:
        vi, vj, = mesh.vertex_ring_vertices_iterators()
    N = vj.shape[0]
    Pj = mesh.vertices[vj] - mesh.vertices[vi]
    x1 = np.einsum('ij,ij -> i', Pj, e1[vi])
    x2 = np.einsum('ij,ij -> i', Pj, e2[vi])
    x3 = np.einsum('ij,ij -> i', Pj, e3[vi])
    i = np.arange(N)
    if interpolate_vertex:
        i = np.hstack((i,i,i,i,i))
        j = 5*vi
        j = np.hstack((j,j+1,j+2,j+3,j+4))
        data = np.hstack((x1, x2, x1**2, x1*x2, x2**2))
        D = coo_matrix((data, (i, j)), shape=(N, 5*mesh.V))
        D = D.tocsr()
        A = (D.T).dot(D)
        b = (D.T).dot(x3)
        a = spsolve(A, b)
        a = np.reshape(a, (mesh.V, 5))
        a = np.column_stack((np.zeros(mesh.V), a))
    else:
        i = np.hstack((i,i,i,i,i,i))
        j = 6*vi
        j = np.hstack((j,j+1,j+2,j+3,j+4,j+5))
        one = np.ones(x1.shape[0])
        data = np.hstack((one, x1, x2, x1**2, x1*x2, x2**2))
        D = coo_matrix((data, (i, j)), shape=(N, 6*mesh.V))
        D = D.tocsr()
        A = (D.T).dot(D)
        b = (D.T).dot(x3)
        a = spsolve(A, b)
        a = np.reshape(a, (mesh.V, 6))
    return a

def principal_curvatures(mesh, double_ring=False, interpolate_vertex=False):
    e3 = mesh.vertex_normals()
    e1 = utilities.orthogonal_vectors(e3)
    e2 = np.cross(e3, e1)
    a = quadratic_fitting(mesh, double_ring, interpolate_vertex)
    g = (a[:,1]**2 + a[:,2]**2) + 1
    W = np.array([[2*g*a[:,3], g*a[:,4]],[g*a[:,4], 2*g*a[:,5]]]).T
    eig = np.linalg.eigh(W)
    k1 = eig[0][:,0]
    k2 = eig[0][:,1]
    V1 = eig[1][:,:,0]
    V2 = eig[1][:,:,1]
    V1 = (e1.T * V1[:,0]).T + (e2.T * V1[:,1]).T
    V2 = (e1.T * V2[:,0]).T + (e2.T * V2[:,1]).T
    return (k1, k2, V1, V2)

'''
def principal_curvatures2(mesh, double_ring=False, interpolate_vertex=True):
    e3 = mesh.vertex_normals()
    e1 = utilities.orthogonal_vectors(e3)
    e2 = np.cross(e3,e1)
    if double_ring:
        vi, vj, = mesh.vertex_double_ring_vertices_iterators()
    else:
        vi, vj, = mesh.vertex_ring_vertices_iterators()
    N = vj.shape[0]
    Pj = mesh.vertices[vj] - mesh.vertices[vi]
    x1 = np.einsum('ij,ij -> i', Pj, e1[vi])
    x2 = np.einsum('ij,ij -> i', Pj, e2[vi])
    x3 = np.einsum('ij,ij -> i', Pj, e3[vi])
    i = np.arange(N)
    if interpolate_vertex:
        i = np.hstack((i,i,i,i,i))
        j = 5*vi
        j = np.hstack((j,j+1,j+2,j+3,j+4))
        data = np.hstack((x1, x2, x1**2, x1*x2, x2**2))
        D = coo_matrix((data, (i, j)), shape=(N, 5*mesh.V))
        D = D.tocsr()
        A = (D.T).dot(D)
        b = (D.T).dot(x3)
        a = spsolve(A, b)
        a1 = a[::5]
        a2 = a[1::5]
        a3 = a[2::5]
        a4 = a[3::5]
        a5 = a[4::5]
    else:
        i = np.hstack((i,i,i,i,i,i))
        j = 6*vi
        j = np.hstack((j,j+1,j+2,j+3,j+4,j+5))
        one = np.ones(x1.shape[0])
        data = np.hstack((one, x1, x2, x1**2, x1*x2, x2**2))
        D = coo_matrix((data, (i, j)), shape=(N, 6*mesh.V))
        D = D.tocsr()
        A = (D.T).dot(D)
        b = (D.T).dot(x3)
        a = spsolve(A, b)
        a1 = a[1::6]
        a2 = a[2::6]
        a3 = a[3::6]
        a4 = a[4::6]
        a5 = a[5::6]
    g = (a1**2 + a2**2) + 1
    W = np.array([[2*g*a3, g*a4],[g*a4, 2*g*a5]]).T
    eig = np.linalg.eigh(W)
    k1 = eig[0][:,0]
    k2 = eig[0][:,1]
    V1 = eig[1][:,:,0]
    V2 = eig[1][:,:,1]
    V1 = (e1.T * V1[:,0]).T + (e2.T * V1[:,1]).T
    V2 = (e1.T * V2[:,0]).T + (e2.T * V2[:,1]).T
    return (k1, k2, V1, V2)
'''


#!/usr/bin/env python

# -*- coding: utf-8 -*-

from __future__ import absolute_import

from __future__ import print_function

from __future__ import division

from traits.api import HasTraits, Instance, Property, Enum, Button, String, \
    on_trait_change, Float, Bool, Int, Constant, ReadOnly, \
    List, Array, Range, Event

from traitsui.api import View, Item, HSplit, VSplit, InstanceEditor, HGroup, \
    Group, ListEditor, Tabbed, VGroup, CheckListEditor, \
    ArrayEditor, Action, ToolBar, Separator, Controller

from tvtk.pyface.scene_editor import SceneEditor

from mayavi.tools.mlab_scene_model import MlabSceneModel

from pyface.image_resource import ImageResource

from pyface.api import FileDialog, OK

from tvtk.pyface.api import Scene

# ------------------------------------------------------------------------------

from wsurf.geolab.gui.scenemanager import SceneManager

from wsurf.geolab.gui import openobj

from wsurf.geolab.gui.multiscenemanager import MultiSceneManager

from wsurf.geolab.plotter.geolabscene import GeolabScene, GeolabSecondaryScene, \
    GeolabBaseScene

from wsurf.geolab.gui.geolabcomponent import GeolabComponent

from wsurf.geolab.gui.tools import CornerTolerance, \
    Loads, SaveGeometry, BsplineRebuilder

# -----------------------------------------------------------------------------

__author__ = 'Davide Pellis'


# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
#                                     Handler
# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------


class GlHandler(Controller):

    def close(self, info, is_ok):
        info.object._closing()
        info.object._closed = True
        Controller.close(self, info, is_ok)
        return True

    # --------------------------------------------------------------------------
    #                                  Tools
    # --------------------------------------------------------------------------

    def open_plot_manager(self):
        self.info.object.open_plot_manager()

    def open_save_geometry(self):
        self.info.object.open_save_geometry()

    def open_import(self):
        self.info.object.open_import()

    def open_bspline_rebuilder(self):
        self.info.object.open_bspline_rebuilder()

    # --------------------------------------------------------------------------
    #                                 Selection
    # --------------------------------------------------------------------------

    def background_switch(self):
        self.info.object.background_switch()

    def select_object(self):
        self.info.object.select_object()

    def select_vertices(self):
        self.info.object.select_vertices()

    def select_boundary_vertices(self):
        self.info.object.select_boundary_vertices()

    def select_all_vertices(self):
        self.info.object.select_all_vertices()

    def move_vertices(self):
        self.info.object.move_vertices()

    # --------------------------------------------------------------------------
    #                                 Mesh state
    # --------------------------------------------------------------------------

    def reset_geometry(self):
        self.info.object.reset_geometry()

    def set_reference(self):
        self.info.object.set_reference()

    def delete_selected_object(self):
        self.info.object.delete_selected_object()

    # --------------------------------------------------------------------------
    #                              Set conditions
    # --------------------------------------------------------------------------

    def fix_vertices(self):
        self.info.object.fix_vertices()

    def unfix_vertices(self):
        self.info.object.unfix_vertices()


# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
#                       The Base Graphic User Interface
# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------


class GeolabGUI(MultiSceneManager):

    _handler = GlHandler()

    _closed = Bool(True)

    _current_object = String('none')

    _components = []

    scene_model_0 = Instance(MlabSceneModel, ())

    scene_model_1 = Instance(MlabSceneModel, ())

    scene_model_2 = Instance(MlabSceneModel, ())

    scene_model_3 = Instance(MlabSceneModel, ())

    components = List(GeolabComponent)

    bspline_rebuilder = BsplineRebuilder()

    corner_tolerance_tool = CornerTolerance()

    loads_tool = Loads()

    save_geometry_tool = SaveGeometry()

    _background_switch_counter = Int(1)

    bspline_rebuilder_button = Action(action='open_bspline_rebuilder',
                                      image=ImageResource('../icons/bspline_rebuild.png'),
                                      style='push',
                                      enabled_when='_current_object == "Bspline_surface_plot_manager"',
                                      tooltip='B-spline rebuilder',
                                      show_label=False)

    background_switch_button = Action(action='background_switch',
                                      image=ImageResource('../icons/background.png'),
                                      style='push',
                                      tooltip='Switch background color',
                                      show_label=False)

    save_geometry_button = Action(action='open_save_geometry',
                                  image=ImageResource('../icons/save.png'),
                                  style='push',
                                  enabled_when='_current_object == "Bspline_surface_plot_manager"',
                                  tooltip='Save geometry',
                                  show_label=False)

    import_button = Action(action='open_import',
                           image=ImageResource('../icons/open.png'),
                           style='push',
                           tooltip='Import geometry',
                           show_label=False)

    plot_manager_button = Action(action='open_plot_manager',
                                 image=ImageResource('../icons/plotmanager.png'),
                                 style='push',
                                 tooltip='Object plot settings',
                                 enabled_when='_current_object == "Bspline_surface_plot_manager"',
                                 show_label=False)

    move_vertices_button = Action(action='move_vertices',
                                  image=ImageResource('../icons/movevertex.png'),
                                  style='toggle',
                                  tooltip='Move vertices',
                                  enabled_when='_current_object == "Bspline_surface_plot_manager"',
                                  show_label=False)

    select_vertices_button = Action(action='select_vertices',
                                    image=ImageResource('../icons/selectvertices.png'),
                                    style='toggle',
                                    tooltip='Select vertices',
                                    enabled_when='_current_object == "Bspline_surface_plot_manager"',
                                    show_label=False)

    select_boundary_vertices_button = Action(action='select_boundary_vertices',
                                             image=ImageResource('../icons/boundary.png'),
                                             style='toggle',
                                             enabled_when='_current_object == "Bspline_surface_plot_manager"',
                                             tooltip='Select boundary vertices',
                                             show_label=False)

    select_all_vertices_button = Action(action='select_all_vertices',
                                        image=ImageResource('../icons/allvertices.png'),
                                        style='toggle',
                                        enabled_when='_current_object == "Bspline_surface_plot_manager"',
                                        tooltip='Select all vertices',
                                        show_label=False)

    fix_vertices_button = Action(action='fix_vertices',
                                 image=ImageResource('../icons/fixvertices.png'),
                                 style='push',
                                 enabled_when='_current_object == "Bspline_surface_plot_manager"',
                                 tooltip='Fix selected vertices',
                                 show_label=False)

    unfix_vertices_button = Action(action='unfix_vertices',
                                   image=ImageResource('../icons/unfixvertices.png'),
                                   style='push',
                                   enabled_when='_current_object == "Bspline_surface_plot_manager"',
                                   tooltip='Unfix selected vertices',
                                   show_label=False)

    reset_geometry_button = Action(action='reset_geometry',
                                   image=ImageResource('../icons/resetmesh.png'),
                                   style='push',
                                   enabled_when=('_current_object == \
                                          "Bspline_surface_plot_manager"'),
                                   tooltip='Reset geometry',
                                   show_label=False)

    set_reference_button = Action(action='set_reference',
                                  image=ImageResource('../icons/setreference.png'),
                                  style='push',
                                  enabled_when='_current_object == "Bspline_surface_plot_manager"',
                                  tooltip='Set as reference geometry',
                                  show_label=False)

    delete_selected_object_button = Action(action='delete_selected_object',
                                           image=ImageResource('../icons/delobject.png'),
                                           style='push',
                                           enabled_when='_current_object == "Bspline_surface_plot_manager"',
                                           tooltip='Delete object',
                                           show_label=False)

    __toolbar = [Separator(),
                 plot_manager_button,

                 Separator(),
                 select_vertices_button,
                 move_vertices_button,
                 select_all_vertices_button,
                 select_boundary_vertices_button,

                 Separator(),
                 fix_vertices_button,
                 unfix_vertices_button,

                 Separator(),
                 bspline_rebuilder_button,

                 Separator(),
                 reset_geometry_button,
                 set_reference_button,
                 delete_selected_object_button,

                 Separator(),
                 import_button,
                 save_geometry_button,

                 Separator(),
                 background_switch_button,
                 ]

    toolbar = {'handler': _handler,
               'resizable': True,
               'title': 'W-Surf',
               'icon': ImageResource('../icons/geolab_logo.png'),
               'toolbar': ToolBar(*__toolbar,
                                  show_labels=False,
                                  image_size=(16, 16)),
               }

    tabs = [Item('components',
                 editor=ListEditor(use_notebook=True, page_name='.name'),
                 style='custom',
                 width=1,
                 resizable=False,
                 show_label=False)
            ]

    __view3D = []

    __windows = []

    height = Int(200)

    width = Int(200)

    side_width = Int(200)

    # -------------------------------------------------------------------------
    #
    # -------------------------------------------------------------------------

    def __init__(self):
        MultiSceneManager.__init__(self)

        self.__scenes = []

        self.__scenes_settings = {}

        self.__geometries = []

        self.__object_open_callbacks = []

        self.__initialize_plot = True

        self.__change = False

        self._scene_model = self.scene_model_0

    # -------------------------------------------------------------------------

    @property
    def geometries(self):
        return self.__geometries

    @property
    def is_moving_vertex(self):
        return self.move_vertices_button.checked

    # -------------------------------------------------------------------------
    #                             Building up
    # -------------------------------------------------------------------------

    def start(self):
        self.make_scenes()
        if len(self._components) > 0 and len(self.__windows) > 0:
            self.components = self._components
            view = View(HSplit(self.tabs,
                               self.__view3D,
                               self.__windows),
                        **self.toolbar)
        elif len(self._components) > 0:
            self.components = self._components
            view = View(HSplit(self.tabs,
                               self.__view3D),
                        **self.toolbar)
        else:
            view = View(self.__view3D, **self.toolbar)
        self._closed = False

        for component in self._components:
            component.initialize_plot()
        for geometry in self.__geometries:
            self.add_object(geometry)
        self.object_changed()

        self.update_plot()

        self.configure_traits(view=view)

    def add_component(self, component):
        component.geolab = self
        self._components.append(component)
        component.geolab_settings()
        self.__initialize_plot = False

    def add_scene(self, name, **kwargs):
        if name not in self.__scenes:
            self.__scenes.append(name)
            self.__scenes_settings[name] = kwargs

    def make_scenes(self):
        if len(self.__scenes) == 0:
            self.__scenes = ['scene_0']
            self.__scenes_settings = {'scene_0': {}}
        index = 0
        for key in self.__scenes:
            if index == 0:
                scene = self.scene_model_0
                self.__view3D = [Item('scene_model_0',
                                      editor=SceneEditor(scene_class=GeolabScene),
                                      show_label=False,
                                      resizable=True,
                                      height=self.height,
                                      width=self.width)]

            else:
                if index == 1:
                    scene = self.scene_model_1
                elif index == 2:
                    scene = self.scene_model_2
                elif index == 3:
                    scene = self.scene_model_3
                else:
                    return
                name = ('scene_model_{}').format(index)
                settings = self.__scenes_settings[key]
                style = settings.get('style', 'geolab')
                if style == 'plain':
                    editor = Scene
                elif style == 'geolab':
                    editor = GeolabSecondaryScene
                elif style == 'base':
                    editor = GeolabBaseScene
                resizable = settings.get('resizable', True)
                self.__windows.append(Item(name,
                                           editor=SceneEditor(scene_class=editor),
                                           show_label=False,
                                           resizable=resizable,
                                           height=100,
                                           width=self.side_width))
            self.add_scene_model(scene, key)
            index += 1
        if index <= 3:
            self.scene_model_3.mayavi_scene.remove()
        if index <= 2:
            self.scene_model_2.mayavi_scene.remove()
        if index <= 1:
            self.scene_model_1.mayavi_scene.remove()

    def _closing(self):
        for component in self._components:
            component.geolab_closing()
        self.scene_model_0.scene.close()
        self.scene_model_1.scene.close()
        self.scene_model_2.scene.close()
        self.scene_model_3.scene.close()

    @on_trait_change('scene_model_0.activated, scene_model_1.activated,\
                     scene_model_2.activated,scene_model_3.activated')
    def _scene_settings(self):
        for name in self.__scenes_settings:
            scene = self.get_scene(name)
            try:
                settings = self.__scenes_settings[name]
                interaction = settings.get('interaction', None)
                position = settings.get('position', None)
                if interaction == '2d':
                    scene.interaction_2d()
                elif interaction == 'locked':
                    scene.interaction_locked()
                if position is not None:
                    scene.position = position
                    scene.set_view(position)
            except:
                pass

    # -------------------------------------------------------------------------
    #                             Standard Methods
    # -------------------------------------------------------------------------

    def set_state(self, name=None):
        self.select_off()
        if name != 'move_vertices':
            self.move_vertices_button.checked = False
        if name != 'select_vertices':
            self.select_vertices_button.checked = False
        if name != 'select_boundary_vertices':
            self.select_boundary_vertices_button.checked = False
        if name != 'select_all_vertices':
            self.select_all_vertices_button.checked = False
        for component in self._components:
            component.geolab_set_state(name)

    def object_change(self):
        SceneManager.object_change(self)

    def object_changed(self, main=True):
        SceneManager.object_changed(self)
        self.close_tools()
        self._current_object = self.current_object_type
        if main:
            self.set_state()

    # --------------------------------------------------------------------------
    #                                Background
    # --------------------------------------------------------------------------

    def background_switch(self):
        if self._background_switch_counter == 1:
            self.set_background((1, 1, 1))
            self._background_switch_counter = 0
        else:
            self.set_background((0.5, 0.5, 0.5))
            self._background_switch_counter = 1

    # -------------------------------------------------------------------------
    #                                  Open
    # -------------------------------------------------------------------------

    def open_obj_file(self, file_name):
        G = openobj.open_obj_file(file_name)
        self.__geometries.append(G)
        if not self._closed:
            empty = False
            if self.current_object is None:
                empty = True
            self.add_object(G)
            if empty:
                self.object_changed()

    def open_geometry(self, geometry):
        self.__geometries.append(geometry)
        if not self._closed:
            self.add_object(geometry)

    def reset_geometry(self):
        self.set_state()
        file_name = self.current_object.geometry.name + '.obj'
        G = openobj.open_obj_file(file_name)
        self.current_object.geometry = G
        self.object_added()
        self.object_changed()
        self.current_object.update_plot()

    def set_reference(self):
        for component in self._components:
            component.geolab_set_reference()

    def delete_selected_object(self):
        self.set_state()
        self.delete_object(self.current_object.name)
        # self._current_object = self.current_object_type
        self.object_changed()

    def add_object(self, obj, name=None, scene=None):
        self.set_state()
        MultiSceneManager.add_object(self, obj, name, scene)

    # --------------------------------------------------------------------------
    #                                  Tools
    # --------------------------------------------------------------------------

    def open_plot_manager(self):
        self.selection_off()
        self.current_object.start()

    def open_bspline_rebuilder(self):
        self.set_state()
        self.bspline_rebuilder.scenemanager = self
        self.bspline_rebuilder.start()

    def open_import(self):
        extensions = ['*.obj']
        descriptions = ["OBJ"]
        wildcard = ""
        for description, extension in zip(descriptions, extensions):
            wildcard += "{} ({})|{}|".format(description,
                                             extension,
                                             extension)
        wildcard += "All (*.*)|(*.*)"
        self.set_state()
        dialog = FileDialog(title='Import geometry', action='open',
                            wildcard=wildcard)
        if dialog.open() == OK:
            self.open_obj_file(dialog.path)

    def open_corner_tolerance(self):
        self.set_state()
        self.corner_tolerance_tool.scenemanager = self
        self.corner_tolerance_tool.start()

    def open_save_geometry(self):
        self.save_geometry_tool.scenemanager = self
        self.save_geometry_tool.start()

    @on_trait_change('_closed')
    def close_all(self):
        self.close_tools()
        self.bspline_rebuilder.close()

    def close_tools(self):
        try:
            self.current_object.close()
        except:
            pass
        self.save_geometry_tool.close()

    # --------------------------------------------------------------------------
    #                                Selection
    # --------------------------------------------------------------------------

    def selection_off(self):
        MultiSceneManager.select_off(self)
        self.select_boundary_vertices_button.checked = False
        self.select_vertices_button.checked = False
        self.move_vertices_button.checked = False

    def select_vertices(self):
        if self.select_vertices_button.checked:
            self.set_state('select_vertices')
            self.current_object.select_vertices()
        else:
            self.current_object.select_vertices_off()

    def select_boundary_vertices(self):
        if self.select_boundary_vertices_button.checked:
            self.set_state('select_boundary_vertices')
            self.current_object.select_boundary_vertices()
        else:
            self.current_object.select_boundary_vertices_off()

    def select_all_vertices(self):
        if self.select_all_vertices_button.checked:

            self.set_state('select_all_vertices')
            self.current_object.select_all_vertices()
        else:
            self.current_object.select_vertices_off()

    def move_vertices(self):
        if self.move_vertices_button.checked:
            self.set_state('move_vertices')
            self.current_object.move_vertices(self.current_object.update_plot)
        else:
            self.current_object.move_vertices_off()

    # --------------------------------------------------------------------------
    #                              Set conditions
    # --------------------------------------------------------------------------

    def fix_vertices(self):
        selected = self.current_object.selected_vertices
        for component in self._components:
            component.geolab_fix_vertices(selected)
        self.current_object.update_plot()

    def unfix_vertices(self):
        selected = self.current_object.selected_vertices
        for component in self._components:
            component.geolab_unfix_vertices(selected)
        self.current_object.update_plot()

# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
#
# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------


class ViewerHandler(Controller):

    def close(self, info, is_ok):
        try:
            info.object._closed = True
            Controller.close(self, info, is_ok)
        except:
            pass
        return True


class Viewer(SceneManager):

    scene = Instance(MlabSceneModel, ())

    _closed = True

    _handler = ViewerHandler()

    def __init__(self, **kwargs):
        SceneManager.__init__(self)

        self._kwargs = kwargs

        self._size = kwargs.get('size', (800, 800))

    @property
    def scene_model(self):
        return self.scene

    def start(self):
        view = View(Item('scene',
                         editor=SceneEditor(scene_class=GeolabScene),
                         show_label=False,
                         resizable=self._kwargs.get('resizable', True),
                         height=self._size[1],
                         width=self._size[0],
                         ),
                    handler=self._handler,
                    resizable=self._kwargs.get('resizable', True),
                    title=self._kwargs.get('title', 'GeoLab'),
                    icon=ImageResource('../icons/geolab_logo.png')
                    )
        self._closed = False
        self.configure_traits(view=view)

    def close(self):
        self._closed = True
        self._handler.info.ui.dispose()

    @on_trait_change('scene.activated')
    def _scene_settings(self):
        position = self._kwargs.get('position', None)
        if position is not None:
            self.set_view(position)
        background = self._kwargs.get('background', None)
        if background is not None:
            self.background = background
        interaction = self._kwargs.get('interaction', None)
        if interaction == '2d':
            self.interaction_2d()
        elif interaction == 'locked':
            self.interaction_locked()


# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
#
# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------




# -*- coding: utf-8 -*-

from __future__ import absolute_import

from __future__ import print_function

from __future__ import division

import numpy as np

# -----------------------------------------------------------------------------

__author__ = 'Davide Pellis'


def linear_regression(points):
    b = np.sum(points, axis=0) / points.shape[0]
    P = np.copy(points)
    for i in range(points.shape[1]):
        P[:, i] -= b[i]
    C = np.einsum('ij,jk->ik', P.T, P)
    eig = np.linalg.eigh(C)
    w = eig[1][:, np.argsort(np.abs(eig[0]))].T
    return w, b


def fit_sphere(points):
    A = np.zeros((len(points), 4))
    A[:, 0] = points[:, 0] * 2
    A[:, 1] = points[:, 1] * 2
    A[:, 2] = points[:, 2] * 2
    A[:, 3] = 1
    f = np.zeros((len(points), 1))
    f[:, 0] = (points[:, 0] ** 2) + (points[:, 1] ** 2) + (points[:, 2] ** 2)
    M = np.dot(A.T, A)
    b = np.dot(A.T, f)
    # C, residules, rank, singval = np.linalg.lstsq(A,f,rcond=-1)
    C = np.linalg.solve(M, b)
    t = (C[0] * C[0]) + (C[1] * C[1]) + (C[2] * C[2]) + C[3]
    radius = np.sqrt(t)
    return radius[0], C[0:3].T



#!/usr/bin/env python

# -*- coding: utf-8 -*-

from __future__ import absolute_import

from __future__ import print_function

from __future__ import division

import numpy as np

from tvtk.api import tvtk

from mayavi.filters.poly_data_normals import PolyDataNormals

from mayavi.sources.vtk_data_source import VTKDataSource

from mayavi.modules.surface import Surface

from mayavi.modules.iso_surface import IsoSurface

# -----------------------------------------------------------------------------

from wsurf.geolab.plotter import plotutilities

# -----------------------------------------------------------------------------

'''facesource.py: The mesh faces plot source class'''

__author__ = 'Davide Pellis'


# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
#                                    Faces
# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------


class Faces(object):

    def __init__(self, mesh, **kwargs):
        plotutilities.check_arguments(**kwargs)

        self._mesh               = mesh

        self._data_range         = None

        self._face_data          = kwargs.get('face_data', None)

        self._vertex_data        = kwargs.get('vertex_data', None)

        self._color              = kwargs.get('color', 'cornflower')

        self._edge_color         = kwargs.get('edge_color', 'black')

        self._opacity            = kwargs.get('opacity', 1)

        self._line_width         = kwargs.get('line_width', 1)

        self._edge_visibility    = kwargs.get('edge_visibility', False)

        self._lut_range          = kwargs.get('lut_range', '-:+')

        self._lut_expansion      = kwargs.get('lut_expansion', 2)

        self._reverse_lut        = kwargs.get('reverse_lut', False)

        self._smooth             = kwargs.get('smooth', True)

        self._shading            = kwargs.get('shading', True)

        self._glossy             = kwargs.get('glossy', 0.3)

        self._specular           = kwargs.get('specular', 1)

        self._iso_surface        = kwargs.get('iso_surface', False)

        self._tube_radius        = kwargs.get('tube_radius', None)

        self._number_of_contours = kwargs.get('number_of_contours', 50)

        self._force_opaque       = kwargs.get('force_opaque', False)

        self._data_points        = kwargs.get('data_points', mesh.vertices)

        self._iso_points         = kwargs.get('iso_points', False)

        self.name                = kwargs.get('name', 'faces')

    # -------------------------------------------------------------------------
    #                                 Pipeline
    # -------------------------------------------------------------------------
    # - .source  =   VTKDataSource
    # - ._data    =     \_Unstructured_grid
    # - ._normals =       None  \_Poly_data_normals
    # - ._module  =        \_______\_Module_manager
    # - ._surface =                   \__Surface
    # -------------------------------------------------------------------------

        self._make_data()

        self._make_normals()

        self._module = plotutilities.make_module(self._color,
                                                 self._opacity,
                                                 self._lut_range,
                                                 self._lut_expansion,
                                                 self._reverse_lut,
                                                 self._data_range,
                                                 points=self._data_points)

        if self._iso_surface or self._iso_points:
            self._make_iso_surface()
        else:
            self._make_surface()

        self._assemble_pipeline()

        self.on_scene = False

    @property
    def type(self):
        return 'Faces-source'

    # -------------------------------------------------------------------------
    #                               Data Structure
    # -------------------------------------------------------------------------

    def _make_data(self):
        cells, types = self._mesh.cell_arrays()
        F = self._mesh.F
        points = self._mesh.vertices
        cell_types = np.array([tvtk.Triangle().cell_type,
                               tvtk.Quad().cell_type,
                               tvtk.Polygon().cell_type])
        cell_types = cell_types[types]
        cells = np.array(cells)
        cell_array = tvtk.CellArray()
        cell_array.set_cells(F, cells)
        self._data = tvtk.UnstructuredGrid(points=points)
        self._data.set_cells(cell_types, [], cell_array)
        if self._face_data is not None:
            scalars = np.array(self._face_data)
            self._data.cell_data.scalars = scalars
            self._data.cell_data.scalars.name = self.name
            self._data_range = [np.min(scalars), np.max(scalars)]
        elif self._vertex_data is not None:
            scalars = np.array(self._vertex_data)
            self._data.point_data.scalars = scalars
            self._data.point_data.scalars.name = self.name
            self._data_range = [np.min(scalars),np.max(scalars)]
        elif self._color == 'xyz' or self._color == 'uv_wheel':
            scalars = np.arange(self._mesh.V)
            self._data.point_data.scalars = scalars
            self._data.point_data.scalars.name = self.name
            self._data_range = [0, np.max(scalars)]
        else:
            scalars = np.zeros([F])
            self._data.cell_data.scalars = scalars
            self._data.cell_data.scalars.name = self.name
            self._data_range = [0, 0]

    # -------------------------------------------------------------------------
    #                                Surface
    # -------------------------------------------------------------------------

    def _make_surface(self):
        self._surface = Surface()
        if self._line_width is not None:
            self._surface.actor.property.line_width = self._line_width
        if self._edge_visibility:
            self._surface.actor.property.edge_visibility = True
            edge_color = plotutilities.map_color(self._edge_color)
            self._surface.actor.property.edge_color = edge_color
        if not self._shading:
            self._surface.actor.actor.property.lighting = False
        if type(self._opacity) == int or type(self._opacity) == float:
            if self._opacity < 1:
                self._surface.actor.actor.property.specular = 0.8
        self._surface.actor.actor.force_opaque = self._force_opaque
        self._surface.actor.actor.property.specular = min(self._glossy, 1)
        sp = 21.*self._specular*self._glossy
        self._surface.actor.actor.property.specular_power = sp + 0.001

    def _make_iso_surface(self):
        self._surface = IsoSurface()
        if self._line_width is not None:
            self._surface.actor.property.line_width = self._line_width
        self._surface.contour.auto_contours = True
        self._surface.contour.number_of_contours = self._number_of_contours
        if self._tube_radius is not None:
            self._surface.actor.property.render_lines_as_tubes = True
            self._surface.actor.property.edge_visibility = True
            self._surface.actor.property.render_points_as_spheres = True
            self._surface.actor.property.line_width = 1 + self._tube_radius
            self._surface.actor.property.point_size = 1 + self._tube_radius
            if self._iso_surface:
                self._surface.actor.property.representation = 'surface'
            else:
                self._surface.actor.property.representation = 'points'
        color = plotutilities.rgb_float_color(self._color)
        self._surface.actor.property.edge_color = color
        self._surface.actor.property.vertex_color = color
        self._surface.actor.actor.property.specular = min(self._glossy, 1)
        sp = 21.*self._specular*self._glossy
        self._surface.actor.actor.property.specular_power = sp + 0.001

    # -------------------------------------------------------------------------
    #                                Normals
    # -------------------------------------------------------------------------

    def _make_normals(self):
        if self._smooth:
            self._normals = PolyDataNormals()
        else:
            self._normals = None

    # -------------------------------------------------------------------------
    #                                Pipeline
    # -------------------------------------------------------------------------

    def _assemble_pipeline(self):
        src = VTKDataSource(data=self._data)
        self._module.add_child(self._surface)
        if self._normals is None:
            src.add_child(self._module)
        else:
            self._normals.add_child(self._module)
            src.add_child(self._normals)
        self.source = src

    # -------------------------------------------------------------------------
    #                                 Update
    # -------------------------------------------------------------------------

    def _update_lut_range(self, **kwargs):
        self._lut_range = kwargs.get('lut_range', self._lut_range)
        lut_range = plotutilities.make_lut_range(self._lut_range, self._data_range)
        self._module.scalar_lut_manager.data_range = lut_range

    def _update_color(self, **kwargs):
        if 'color' in kwargs:
            color = kwargs['color']
            if color is not self._color:
                self._color = kwargs['color']
                plotutilities.make_lut_table(self._module,
                                             self._color,
                                             self._opacity,
                                             self._lut_expansion,
                                             self._reverse_lut,
                                             points = self._mesh.vertices)

    def _update_data(self, **kwargs):
        if True:
            vertices = kwargs.get('vertices', None)
            if vertices is None:
                vertices = self._mesh.vertices
            self._data.set(points = vertices)
            cells, types = self._mesh.cell_arrays()
            cell_types = np.array([tvtk.Triangle().cell_type,
                                   tvtk.Quad().cell_type,
                                   tvtk.Polygon().cell_type])
            cell_types = cell_types[types]
            cells = np.array([cells])
            cell_array = tvtk.CellArray()
            cell_array.set_cells(self._mesh.F, cells)
            self._data.set_cells(cell_types, [], cell_array)
        self._face_data = kwargs.get('face_data', self._face_data)
        self._vertex_data = kwargs.get('vertex_data', self._vertex_data)
        if self._face_data is not None:
            scalars = np.array(self._face_data)
            self._data.cell_data.scalars = scalars
            self._data.cell_data.scalars.name = self.name
            self._data_range = [np.min(scalars), np.max(scalars)]
        elif self._vertex_data is not None:
            scalars = np.array(self._vertex_data)
            self._data.point_data.scalars = scalars
            self._data.point_data.scalars.name = self.name
            self._data_range = [np.min(scalars), np.max(scalars)]
        elif self._color == 'xyz':
            scalars = np.arange(self._mesh.V)
            self._data.point_data.scalars = scalars
            self._data.point_data.scalars.name = self.name
            self._data_range = [np.min(scalars), np.max(scalars)]
        else:
            scalars = np.zeros([self._mesh.F])
            self._data.cell_data.scalars = scalars
            self._data.cell_data.scalars.name = self.name
            self._data_range = [0, 0]
        self._data.modified()

    def _update_glossy(self, **kwargs):
        self._glossy = kwargs.get('glossy', self._glossy)
        self._surface.actor.actor.property.specular =  min(self._glossy, 1)
        sp = 21.*self._specular*self._glossy
        self._surface.actor.actor.property.specular_power = sp + 0.001

    # -------------------------------------------------------------------------
    #
    # -------------------------------------------------------------------------

    def update(self, **kwargs):
        self._update_data(**kwargs)
        self._update_lut_range(**kwargs)
        self._update_color(**kwargs)
        self._update_glossy(**kwargs)
        self.source.update()

    def kwargs(self):
        kwargs = {}
        kwargs['name'] = self.name
        kwargs['vertices'] = np.copy(self._mesh.vertices)
        kwargs['vertex_data'] = np.copy(self._vertex_data)
        kwargs['face_data'] = np.copy(self._face_data)
        return kwargs

    # -------------------------------------------------------------------------
    #                                Legend
    # -------------------------------------------------------------------------

    def show_legend(self, **kwargs):
        label = kwargs.get('label', 'faces')
        self._module.scalar_lut_manager.show_legend = True
        self._module.scalar_lut_manager.data_name = label
        #print(dir(self._module.scalar_lut_manager.scalar_bar))

    def hide_legend(self):
        self._module.scalar_lut_manager.show_legend = False




# -*- coding: utf-8 -*-

from __future__ import absolute_import

from __future__ import print_function

from __future__ import division

import sys

import os

# -----------------------------------------------------------------------------

'''_'''

__author__ = 'Davide Pellis'

# -----------------------------------------------------------------------------


def print_progress(iteration, max_iteration):
    out = '|' + '#'*iteration + '.'*(max_iteration-iteration) + '|\r'
    sys.stdout.write(out)
    sys.stdout.flush()


def make_filepath(file_name, ext='txt', overwrite=False):
    folder = os.getcwd()
    name = os.path.join(folder, file_name)
    path = ('{}.{}').format(name, ext)
    if not overwrite:
        n = 1
        while os.path.exists(path):
            path = ('{}_({}).{}').format(name, n, ext)
            n += 1
    return path


def serial(n, digits=3):
    num = str(n)
    l = len(num)
    s = max(digits, l)
    for i in range(s-l):
        num = '0' + num
    return num



from wsurf.geolab.optimization.sparse_mkl._mkl_interface import (MKL, _sanity_check, _empty_output_check, _type_check, _create_mkl_sparse,
                                           _destroy_mkl_handle, matrix_descr, debug_print, _convert_to_csr,
                                           _get_numpy_layout, _check_return_value, LAYOUT_CODE_C, LAYOUT_CODE_F,
                                           _out_matrix)
import numpy as np
import ctypes as _ctypes
import scipy.sparse as _spsparse


def _sparse_dense_matmul(matrix_a, matrix_b, scalar=1., transpose=False, out=None, out_scalar=None, out_t=None):
    """
    Multiply together a sparse and a dense matrix
    mkl_sparse_?_mm requires the left (A) matrix to be sparse and the right (B) matrix to be dense
    This requires conversion of the sparse matrix to CSR format for some dense arrays.
    A must be CSR if B is column-major. Otherwise CSR or CSC are acceptable.

    :param matrix_a: Left (A) matrix
    :type matrix_a: sp.spmatrix.csr, sp.spmatrix.csc
    :param matrix_b: Right (B) matrix
    :type matrix_b: np.ndarray
    :param scalar: A value to multiply the result matrix by. Defaults to 1.
    :type scalar: float
    :param transpose: Return AT (dot) B instead of A (dot) B.
    :type transpose: bool
    :param out: Add the dot product to this array if provided.
    :type out: np.ndarray, None
    :param out_scalar: Multiply the out array by this scalar if provided.
    :type out_scalar: float, None
    :return: A (dot) B as a dense array in either column-major or row-major format
    :rtype: np.ndarray
    """

    output_shape = (matrix_a.shape[1] if transpose else matrix_a.shape[0], matrix_b.shape[1])
    layout_b, ld_b = _get_numpy_layout(matrix_b, second_arr=out)

    # Prep MKL handles and check that matrixes are compatible types
    # MKL requires CSR format if the dense array is column-major
    if layout_b == LAYOUT_CODE_F and not _spsparse.isspmatrix_csr(matrix_a):
        mkl_non_csr, dbl = _create_mkl_sparse(matrix_a)
        mkl_a = _convert_to_csr(mkl_non_csr)
    else:
        mkl_a, dbl = _create_mkl_sparse(matrix_a)

    # Set functions and types for float or doubles
    output_ctype = _ctypes.c_double if dbl else _ctypes.c_float
    output_dtype = np.float64 if dbl else np.float32
    func = MKL._mkl_sparse_d_mm if dbl else MKL._mkl_sparse_s_mm

    # Allocate an output array
    output_arr = _out_matrix(output_shape, output_dtype, order="C" if layout_b == LAYOUT_CODE_C else "F",
                             out_arr=out, out_t=out_t)

    _, output_ld = _get_numpy_layout(output_arr)

    ret_val = func(11 if transpose else 10,
                   scalar,
                   mkl_a,
                   matrix_descr(),
                   layout_b,
                   matrix_b,
                   output_shape[1],
                   ld_b,
                   float(out_scalar) if out_scalar is not None else 1.,
                   output_arr.ctypes.data_as(_ctypes.POINTER(output_ctype)),
                   output_ld)

    # Check return
    _check_return_value(ret_val, func.__name__)

    _destroy_mkl_handle(mkl_a)

    return output_arr


def _sparse_dot_dense(matrix_a, matrix_b, cast=False, scalar=1., out=None, out_scalar=None):
    """
    Multiply together a dense and a sparse matrix.
    If the sparse matrix is not CSR, it may need to be reordered, depending on the order of the dense array.

    :param matrix_a: Left (A) matrix
    :type matrix_a: np.ndarray, sp.spmatrix.csr, sp.spmatrix.csc
    :param matrix_b: Right (B) matrix
    :type matrix_b: np.ndarray, sp.spmatrix.csr, sp.spmatrix.csc
    :param scalar: A value to multiply the result matrix by. Defaults to 1.
    :type scalar: float
    :param cast: Convert values to compatible floats if True. Raise an error if they are not compatible if False.
    Defaults to False.
    :type cast: bool
    :param out: Add the dot product to this array if provided.
    :type out: np.ndarray, None
    :param out_scalar: Multiply the out array by this scalar if provided.
    :type out_scalar: float, None

    :return: A (dot) B as a dense matrix
    :rtype: np.ndarray
    """
    _sanity_check(matrix_a, matrix_b)

    # Check for edge condition inputs which result in empty outputs
    if _empty_output_check(matrix_a, matrix_b):
        debug_print("Skipping multiplication because A (dot) B must yield an empty matrix")
        final_dtype = np.float64 if matrix_a.dtype != matrix_b.dtype or matrix_a.dtype != np.float32 else np.float32
        return _out_matrix((matrix_a.shape[0], matrix_b.shape[1]), final_dtype, out_arr=out)

    matrix_a, matrix_b = _type_check(matrix_a, matrix_b, cast=cast)

    if sum([_spsparse.isspmatrix(matrix_a), _spsparse.isspmatrix(matrix_b)]) != 1:
        raise ValueError("_sparse_dot_dense takes one sparse and one dense array")
    elif _spsparse.isspmatrix(matrix_a):
        return _sparse_dense_matmul(matrix_a, matrix_b, scalar=scalar, out=out, out_scalar=out_scalar)
    elif _spsparse.isspmatrix(matrix_b) and out is not None:
        _ = _sparse_dense_matmul(matrix_b, matrix_a.T, scalar=scalar, transpose=True,
                                 out=out.T, out_scalar=out_scalar, out_t=True)
        return out
    elif _spsparse.isspmatrix(matrix_b) and out is None:
        return _sparse_dense_matmul(matrix_b, matrix_a.T, scalar=scalar, transpose=True).T




from __future__ import absolute_import

from __future__ import print_function

from __future__ import division

from io import open

# -----------------------------------------------------------------------------

from wsurf.geolab.geometry.bsplinesurface import BsplineSurface


def open_obj_file(file_name):
    file_name = str(file_name)
    obj_file = open(file_name, encoding='utf-8')
    for l in obj_file:
        splited_line = l.split(' ')
        # bspline surface
        if splited_line[0] == 'surf':
           return BsplineSurface(file_name)
    return None


# -*- coding: utf-8 -*-


# ----------------------------------- classes ---------------------------------

from wsurf.geolab.plotter.geolabscene import GeolabScene

from wsurf.geolab.plotter.plotmanager import PlotManager

# -------------------------------- plot classes -------------------------------

from wsurf.geolab.plotter.facesource import Faces

from wsurf.geolab.plotter.edgesource import Edges

from wsurf.geolab.plotter.pointsource import Points

from wsurf.geolab.plotter.polylinesource import Polyline

from wsurf.geolab.plotter.vectorsource import Vectors

# --------------------------------- functions ---------------------------------

from wsurf.geolab.plotter.viewer import view

from wsurf.geolab.plotter import plotutilities

from wsurf.geolab.plotter import colorutilities





""" A viewer for GeoLab scene. Adds a button to open up the engine.
"""

# Standard library imports
from os.path import join

import numpy as np

# Enthought library imports
from tvtk.tools.ivtk import IVTK
from tvtk.pyface.api import DecoratedScene
from traits.api import Callable, Int
from pyface.api import ImageResource
from pyface.action.api import Action, Group
from pyface.resource.api import resource_path

# Local imports
from mayavi.core.common import error
from mayavi.preferences.api import set_scene_preferences, \
        get_scene_preferences


###############################################################################
# A decorated scene with an additional button.
###############################################################################
class GeolabScene(DecoratedScene):

    image_search_path = [join(resource_path(), 'images'), ]

    _position = None

    view_set = Int(1)

    ##########################################################################
    # Non-public interface.
    ##########################################################################
    def show_engine(self):
        """ Open the engine view corresponding to the engine of the
            scene.
        """
        from mayavi.core.registry import registry
        from mayavi.core.ui.engine_rich_view import EngineRichView
        try:
            engine = registry.find_scene_engine(self)
        except TypeError:
            error('This scene is not managed by Mayavi')
        return EngineRichView(engine=engine).scene_editing_view(scene=self)

    ######################################################################
    # Trait handlers.
    ######################################################################
    def _actions_default(self):
        actions = [ Group(
                        Action(tooltip="View the pipeline",
                               image=ImageResource('../icons/mayavi.png'),
                               on_perform=self.show_engine,
                               ),
                         ),
                    Group(
                        Action(tooltip="Save current view",
                               image=ImageResource('../icons/saveview.png'),
                               on_perform=self._save_view,
                               ),
                        Action(tooltip="Apply saved view",
                               image=ImageResource('../icons/setview.png'),
                               on_perform=self._set_view,
                               ),
                        ),
                        ]
        actions.extend(DecoratedScene._actions_default(self))
        return actions

    def _save_view(self):
        self._position = self._get_position()
        print('position = ' + str(self._position))
        self.view_set *= -1

    def _get_position(self):
        cc = self.camera
        p = [cc.position[0],cc.position[1],cc.position[2]]
        p.extend([cc.view_up[0],cc.view_up[1],cc.view_up[2]])
        p.extend([cc.focal_point[0],cc.focal_point[1],cc.focal_point[2]])
        p.extend([cc.view_angle])
        p.extend([cc.clipping_range[0],cc.clipping_range[1]])
        p.extend([self.parallel_projection])
        return p

    def _set_view(self, position=None):
        cc = self.camera
        p = self._position
        if position is not None:
            p = position
        if p is not None:
            cc.position = np.array([p[0],p[1],p[2]])
            cc.view_up = np.array([p[3],p[4],p[5]])
            cc.focal_point = np.array([p[6],p[7],p[8]])
            cc.view_angle = np.float(p[9])
            cc.clipping_range = np.array([p[10],p[11]])
            try:
                cc.parallel_projection = p[12]
            except:
                pass
            self.render()

def mayavi_scene_factory(parent):
    """A mayavi scene factory that creates a scene with preferences
    appropriately set."""
    p = get_scene_preferences()
    s = GeolabScene(parent, stereo=p['stereo'])
    set_scene_preferences(s, p)
    return s


###############################################################################
# A viewer making use of the MayaviScene
###############################################################################
class GeolabViewer(IVTK):
    """ A viewer window for mlab.
    """

    _scene_factory = Callable(mayavi_scene_factory)

    def _size_default(self):
        return (600, 300)


def viewer_factory(size=(400, 350)):
    viewer = GeolabViewer()
    viewer.menu_bar_manager = None
    viewer.size=size
    viewer.open()
    return viewer

if __name__ == '__main__':
    from mayavi.tools.show import show
    viewer_factory()
    show()

###############################################################################
# A decorated scene.
###############################################################################
class GeolabSecondaryScene(DecoratedScene):

    _position = None

    ######################################################################
    # Trait handlers.
    ######################################################################
    def _actions_default(self):
        actions = [ Group(
                        Action(tooltip="Save current view",
                               image=ImageResource('../icons/saveview.png'),
                               on_perform=self._save_view,
                               ),
                        Action(tooltip="Apply saved view",
                               image=ImageResource('../icons/setview.png'),
                               on_perform=self._set_view,
                               ),
                        ),
                        ]
        actions.extend(DecoratedScene._actions_default(self))
        return actions

    def _save_view(self):
        self._position = self._get_position()
        print('position = ' + str(self._position))

    def _get_position(self):
        cc = self.camera
        p = [cc.position[0],cc.position[1],cc.position[2]]
        p.extend([cc.view_up[0],cc.view_up[1],cc.view_up[2]])
        p.extend([cc.focal_point[0],cc.focal_point[1],cc.focal_point[2]])
        p.extend([cc.view_angle])
        p.extend([cc.clipping_range[0],cc.clipping_range[1]])
        p.extend([self.parallel_projection])
        return p

    def _set_view(self, position=None):
        cc = self.camera
        p = self._position
        if position is not None:
            p = position
        if p is not None:
            cc.position = np.array([p[0],p[1],p[2]])
            cc.view_up = np.array([p[3],p[4],p[5]])
            cc.focal_point = np.array([p[6],p[7],p[8]])
            cc.view_angle = np.float(p[9])
            cc.clipping_range = np.array([p[10],p[11]])
            try:
                cc.parallel_projection = p[12]
            except:
                pass
            self.render()

###############################################################################
# A decorated scene.
###############################################################################
class GeolabBaseScene(DecoratedScene):

    ######################################################################
    # Trait handlers.
    ######################################################################
    def _actions_default(self):
        actions = DecoratedScene._actions_default(self)
        return actions[-1:]




#!/usr/bin/env python

# -*- coding: utf-8 -*-

from __future__ import absolute_import

from __future__ import print_function

from __future__ import division

from mayavi.api import Engine

# -----------------------------------------------------------------------------

from wsurf.geolab.gui.scenemanager import SceneManager

# -----------------------------------------------------------------------------

'''plotmanager.py: The plot manager class'''

__author__ = 'Davide Pellis'

# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
#                            Multiple Scene Manager
# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------


class MultiSceneManager(SceneManager):

    def __init__(self):
        SceneManager.__init__(self)

        self._scene_managers = {}

        self._object_scenes = {}

        self._scene_register = {}

    @property
    def object_keys(self):
        return list(self._object_scenes.keys())

    @property
    def scene_keys(self):
        return list(self._scene_managers.keys())

    @property
    def scene_managers(self):
        return self._scene_managers

    @property
    def current_object(self):
        if self._selected is None:
            try:
                return self.get_object(self.object_keys[0])
            except:
                return None
        else:
            return self.get_object(self._selected)

    @property
    def last_object(self):
        if self._last_object is None:
            try:
                return self.get_object(self.object_keys[0])
            except:
                return None
        else:
            return self.get_object(self._last_object)

    @property
    def current_object_type(self):
        if self._selected is None:
            try:
                return self.get_object(self.object_keys[0]).type
            except:
                return 'none'
        else:
            return self.get_object(self._selected).type


    #--------------------------------------------------------------------------
    #                                 Scenes
    #--------------------------------------------------------------------------

    def add_scene_model(self, scene_model, name):
        manager = SceneManager(scene_model)
        self._scene_managers[name] = manager
        self._scene_register[scene_model.mayavi_scene] = manager

    def get_scene(self, name):
        return self._scene_managers[name]

    def get_object_scene(self, name):
        scene = self.get_scene(self._object_scenes[name])
        return scene

    def _format_scene(self, scene):
        if type(scene) is dict:
            scene = scene.get('scene', None)
        if scene is None:
            scene = list(self._scene_managers.keys())[0]
        return scene

    def set_background(self, color):
        for key in self._scene_managers:
            self._scene_managers[key].background = color

    # -------------------------------------------------------------------------
    #                                 Objects
    # -------------------------------------------------------------------------

    def delete_object(self, name):
        self.remove(name)
        scene = self.get_object_scene(name)
        scene.remove(name)
        del self._object_scenes[name]

    def add_object(self, geometry, name=None, scene=None):
        self.object_add()
        if name is None:
            name = 'obj_{}'.format(self._counter)
            self._counter += 1
        scene = self._format_scene(scene)
        self.get_scene(scene).add_object(geometry, name=name)
        self._last_object = name
        obj = self.get_scene(scene).get_object(name)
        self._object_scenes[name] = scene
        self._objects[name] = obj
        for key in self._objects:
            other_obj = self._objects[key]
            if obj != other_obj:
                if (id(other_obj.geometry.vertices) ==
                                        id(obj.geometry.vertices)):
                    obj.add_cross_callback(other_obj.update_plot, name=name)
                    other_obj.add_cross_callback(obj.update_plot, name=name)
        self.object_added()

    def add(self, objects, scene=None, rename=False):
        scene = self._format_scene(scene)
        self.get_scene(scene).add(objects, rename=rename)
        for scene in self.scene_managers:
            for key, obj in self.scene_managers[scene]._objects.items():
                self._object_scenes[key] = scene
                self._objects[key] = obj

    def get_object(self, name):
        scene = self.get_scene(self._object_scenes[name])
        obj = scene.get_object(name)
        return obj

    def clear(self, scene=None):
        scene = self._format_scene(scene)
        self.get_scene(scene).clear()

    def hide(self, name):
        for key in self.scene_managers:
            self.get_scene(key).hide(name)

    # -------------------------------------------------------------------------
    #                                 Plot
    # -------------------------------------------------------------------------

    def disable_render(self):
        for key in self.scene_managers:
            self.get_scene(key).disable_render()

    def enable_render(self):
        for key in self.scene_managers:
            self.get_scene(key).enable_render()

    def plot_points(self, points, **kwargs):
        scene = self._format_scene(kwargs)
        self.get_scene(scene).plot_points(points, **kwargs)

    def plot_polyline(self, polyline, **kwargs):
        scene = self._format_scene(kwargs)
        self.get_scene(scene).plot_polyline(polyline, **kwargs)

    def plot_mesh_faces(self, mesh, **kwargs):
        scene = self._format_scene(kwargs)
        self.get_scene(scene).plot_mesh_faces(mesh, **kwargs)

    def plot_mesh_edges(self, mesh, **kwargs):
        scene = self._format_scene(kwargs)
        self.get_scene(scene).plot_mesh_edges(mesh, **kwargs)

    def update_plot(self, scene=None):
        self.disable_render()
        if scene == None:
            for key in self._scene_managers:
                self.get_scene(key).update_plot()
        else:
            scene = self._format_scene(scene)
            self.get_scene(scene).update_plot()
        for key in self._plot_callbacks:
            self._plot_callbacks[key]()
        self.enable_render()

    # -------------------------------------------------------------------------
    #                                 Select
    # -------------------------------------------------------------------------

    def select_object(self):
        if self.current_object is None:
            return
        self.current_object.highlight()
        ranges = {}
        count = 0
        for key in self._objects:
            obj = self._objects[key]
            N = obj.selection_on()
            ranges[key] = [count, count+N]
            count += N
        if self._selected is None:    
            self._selected = list(self._objects.keys())[0]
        def picker_callback(picker):
            pick_id = picker.cell_id
            if pick_id >= 0:
                for key in self._objects:
                    try:
                        obj = self._objects[key]
                        if picker.actor in obj.object_selection_actors:
                            if key != self._selected:
                                self.current_object.highlight_off()
                                self.object_change()
                                self._selected = key
                                self.object_changed()
                                self.current_object.highlight()
                    except IndexError:
                        pass
        for key in self.scene_managers:
            s = self.scene_managers[key].scene_model.mayavi_scene
            p = s._mouse_pick_dispatcher.callbacks
            if len(p) == 0:
                s.on_mouse_pick(picker_callback, 'cell')
            else:
                p[0] = (picker_callback, 'cell', 'Left')
            a = s._mouse_pick_dispatcher._active_pickers['cell']
            a.tolerance = self.picker_tolerance

    def select_object_off(self):
        self.select_off()

    def select_off(self):
        for key in self._objects:
            obj = self._objects[key]
            obj.select_off()
            obj.highlight_off()
        def picker_callback(picker):
            return
        for key in self.scene_managers:
            s = self.scene_managers[key].scene_model.mayavi_scene
            p = s._mouse_pick_dispatcher.callbacks
            for i in range(len(p)):
                p[i] = (picker_callback, 'cell', 'Left')






from wsurf.geolab.optimization.sparse_mkl._sparse_sparse import _sparse_dot_sparse as _sds
from wsurf.geolab.optimization.sparse_mkl._sparse_dense import _sparse_dot_dense as _sdd
from wsurf.geolab.optimization.sparse_mkl._dense_dense import _dense_dot_dense as _ddd
from wsurf.geolab.optimization.sparse_mkl._sparse_vector import _sparse_dot_vector as _sdv
from wsurf.geolab.optimization.sparse_mkl._gram_matrix import _gram_matrix as _gm
from wsurf.geolab.optimization.sparse_mkl._sparse_qr_solver import sparse_qr_solver as _qrs
from wsurf.geolab.optimization.sparse_mkl._mkl_interface import print_mkl_debug, _is_dense_vector, set_debug_mode, get_version_string
import scipy.sparse as _spsparse
import numpy as _np
import warnings


def dot_product_mkl(matrix_a, matrix_b, cast=False, copy=True, reorder_output=False, dense=False, debug=False,
                    out=None, out_scalar=None):
    """
    Multiply together matrixes using the intel Math Kernel Library.
    This currently only supports float32 and float64 data

    :param matrix_a: Sparse matrix A in CSC/CSR format or dense matrix in numpy format
    :type matrix_a: scipy.sparse.spmatrix, np.ndarray
    :param matrix_b: Sparse matrix B in CSC/CSR format or dense matrix in numpy format
    :type matrix_b: scipy.sparse.spmatrix, np.ndarray
    :param cast: Should the data be coerced into float64 if it isn't float32 or float64
    If set to True and any other dtype is passed, the matrix data will copied internally before multiplication
    If set to False and any dtype that isn't float32 or float64 is passed, a ValueError will be raised
    Defaults to False
    :param copy: Deprecated flag to force copy. Removed because the behavior was inconsistent.
    :type copy: bool
    :param reorder_output: Should the array indices be reordered using MKL
    If set to True, the object in C will be ordered and then exported into python
    If set to False, the array column indices will not be ordered.
    The scipy sparse dot product does not yield ordered column indices so this defaults to False
    :type reorder_output: bool
    :param dense: Should the matrix multiplication be put into a dense numpy array
    This does not require any copy and is memory efficient if the output array density is > 50%
    Note that this flag has no effect if one input array is dense; then the output will always be dense
    :type dense: bool
    :param debug: Deprecated debug flag. Use `sparse_dot_mkl.set_debug_mode(True)`
    :type debug: bool
    :param out: Add the dot product to this array if provided.
    :type out: np.ndarray, None
    :param out_scalar: Multiply the out array by this scalar if provided.
    :type out_scalar: float, None
    :return: Matrix that is the result of A * B in input-dependent format
    :rtype: scipy.sparse.csr_matrix, scipy.sparse.csc_matrix, np.ndarray
    """

    warnings.warn("Set debug mode with sparse_dot_mkl.set_debug_mode(True)", DeprecationWarning) if debug else None
    print_mkl_debug()

    num_sparse = sum((_spsparse.issparse(matrix_a), _spsparse.issparse(matrix_b)))

    # SPARSE (DOT) SPARSE #
    if num_sparse == 2 and out is not None:
        raise ValueError("out argument cannot be used with sparse (dot) sparse matrix multiplication")

    elif num_sparse == 2:
        return _sds(matrix_a, matrix_b, cast=cast, reorder_output=reorder_output, dense=dense)

    # SPARSE (DOT) VECTOR #
    elif num_sparse == 1 and _is_dense_vector(matrix_a) and (matrix_a.ndim == 1 or matrix_a.shape[0] == 1):
        return _sdv(matrix_a, matrix_b, cast=cast, out=out, out_scalar=out_scalar)

    # SPARSE (DOT) VECTOR #
    elif num_sparse == 1 and _is_dense_vector(matrix_b) and (matrix_b.ndim == 1 or matrix_b.shape[1] == 1):
        return _sdv(matrix_a, matrix_b, cast=cast, out=out, out_scalar=out_scalar)

    # SPARSE (DOT) DENSE & DENSE (DOT) SPARSE #
    elif num_sparse == 1:
        return _sdd(matrix_a, matrix_b, cast=cast, out=out, out_scalar=out_scalar)

    # SPECIAL CASE OF VECTOR (DOT) VECTOR #
    # THIS IS JUST EASIER THAN GETTING THIS EDGE CONDITION RIGHT IN MKL #
    elif _is_dense_vector(matrix_a) and _is_dense_vector(matrix_b) and (matrix_a.ndim == 1 or matrix_b.ndim == 1):
        if out_scalar is not None:
            out *= out_scalar
        return _np.dot(matrix_a, matrix_b, out=out)

    # DENSE (DOT) DENSE
    else:
        return _ddd(matrix_a, matrix_b, cast=cast, out=out, out_scalar=out_scalar)


def gram_matrix_mkl(matrix, transpose=False, cast=False, dense=False, debug=False, reorder_output=False,
                    out=None, out_scalar=None):
    """
    Calculate a gram matrix (AT (dot) A) matrix.
    Note that this should calculate only the upper triangular matrix.
    However providing a sparse matrix with transpose=False and dense=True will calculate a full matrix
    (this appears to be a bug in mkl_sparse_?_syrkd)

    :param matrix: Sparse matrix in CSR or CSC format or numpy array
    :type matrix: scipy.sparse.csr_matrix, scipy.sparse.csc_matrix, numpy.ndarray
    :param transpose: Calculate A (dot) AT instead
    :type transpose: bool
    :param cast: Make internal copies to convert matrix to a float matrix or convert to a CSR matrix if necessary
    :type cast: bool
    :param dense: Produce a dense matrix output instead of a sparse matrix
    :type dense: bool
    :param debug: Deprecated debug flag. Use `sparse_dot_mkl.set_debug_mode(True)`
    :type debug: bool
    :param reorder_output: Should the array indices be reordered using MKL
    The scipy sparse dot product does not yield ordered column indices so this defaults to False
    :type reorder_output: bool
    :param out: Add the dot product to this array if provided.
    :type out: np.ndarray, None
    :param out_scalar: Multiply the out array by this scalar if provided.
    :type out_scalar: float, None
    :return: Gram matrix
    :rtype: scipy.sparse.csr_matrix, np.ndarray"""

    warnings.warn("Set debug mode with sparse_dot_mkl.set_debug_mode(True)", DeprecationWarning) if debug else None
    print_mkl_debug()

    return _gm(matrix, transpose=transpose, cast=cast, dense=dense, reorder_output=reorder_output,
               out=out, out_scalar=out_scalar)


def sparse_qr_solve_mkl(matrix_a, matrix_b, cast=False, debug=False):
    """
    Solve AX = B for X where A is sparse and B is dense

    :param matrix_a: Sparse matrix (solver requires CSR; will convert if cast=True)
    :type matrix_a: np.ndarray
    :param matrix_b: Dense matrix
    :type matrix_b: np.ndarray
    :param cast: Should the data be coerced into float64 if it isn't float32 or float64,
    and should a CSR matrix be cast to a CSC matrix.
    Defaults to False
    :type cast: bool
    :param debug: Deprecated debug flag. Use `sparse_dot_mkl.set_debug_mode(True)`
    :type debug: bool
    :return: Dense array X
    :rtype: np.ndarray
    """

    warnings.warn("Set debug mode with sparse_dot_mkl.set_debug_mode(True)", DeprecationWarning) if debug else None
    print_mkl_debug()

    return _qrs(matrix_a, matrix_b, cast=cast)


# Alias for backwards compatibility
dot_product_transpose_mkl = gram_matrix_mkl
